"""
interferometry.py is a Python routine that computes the response and
sensitivity functions of interferometer space-based GW detectors, e.g., LISA and Taiji,
to the detection of SGWB (see tutorial on interferometry/interferometry.ipynb).

Adapted from the original interferometry in GW_turbulence
(https://github.com/AlbertoRoper/GW_turbulence),
created in May 2022

Currently part of the cosmoGW code:

https://github.com/cosmoGW/cosmoGW/
https://github.com/cosmoGW/cosmoGW/blob/main/src/cosmoGW/interferometry.py

Author:  Alberto Roper Pol
Created: 01/05/2022
Updated: 01/11/2023 (preparation of cosmoGW code, included tutorial)
Updated: 18/07/2025 (release cosmoGW 1.0: https://pypi.org/project/cosmoGW)

Main references are:

Caprini:2019pxz  - C. Caprini, D. Figueroa, R. Flauger, G. Nardini, M. Peloso,
M. Pieroni, A. Ricciardone, G. Tassinato, "Reconstructing the spectral shape
of a stochastic gravitational wave background with LISA," JCAP 11 (2019), 017,
arXiv:1906.09244

Schmitz:2020syl  - K. Schmitz "New Sensitivity Curves for Gravitational-Wave Signals from
Cosmological Phase Transitions," JHEP 01, 097 (2021), arXiv:2002.04615.

Orlando:2020oko  - G. Orlando, M. Pieroni, A. Ricciardone, "Measuring Parity Violation
in the Stochastic Gravitational Wave Background with the LISA-Taiji network,"
JCAP 03, 069 (2021), arXiv:2011.07059.

RoperPol:2021xnd - A. Roper Pol, S. Mandal, A. Brandenburg, T. Kahniashvili,
"Polarization of gravitational waves from helical MHD turbulent sources,"
JCAP 04 (2022), 019, arXiv:2107.05356.

RoperPol:2022iel - A. Roper Pol, C. Caprini, A. Neronov,
D. Semikoz, "The gravitational wave signal from primordial
magnetic fields in the Pulsar Timing Array frequency band,"
Phys. Rev. D 105, 123502 (2022), arXiv:2201.05630
"""

import astropy.constants as const
import astropy.units     as u
import numpy  as np
import pandas as pd
import cosmoGW.cosmology as co

from cosmoGW.cosmoGW import HOME
dir0 = HOME + 'resources/detectors_sensitivity/'

# Reference values for LISA and Taiji interferometers
L_LISA   = 2.5e6*u.km
P_LISA   = 15
A_LISA   = 3
L_Taiji  = 3e6*u.km
P_Taiji  = 8
A_Taiji  = 3
SNR_PLS  = 10
T_PLS    = 4
v_dipole = 1.23e-3

# Reference f and beta arrays used in the calculations
f_ref    = np.logspace(-4, 0, 5000)*u.Hz
beta_ref = np.linspace(-20, 20, 3000)

################## SENSITIVITIES AND NOISE POWER SPECTRAL DENSITY ##################

################### READING FUNCTIONS FROM FILES ON SENSITIVITY ###################

def read_response_LISA_Taiji(dir0=dir0, TDI=True, interf='LISA'):

    """
    Function that reads the response functions of LISA and Taiji from the files
    'dir0/LISA_response_f.csv' and 'dir0/Taiji_response_f.csv',
    generated and saved in the routine 'compute_response_LISA_Taiji'.

    TDI chanels are defined following RoperPol:2021xnd, appendix B.

    Arguments:
        dir0 -- directory where to save the results (default is 'detector_sensitivity')
        TDI  -- option to read the response functions for TDI chanels, instead of XYZ
                chanels (default True)
    """

    if TDI:
        dirr = dir0 + interf + '_response_f_TDI.csv'
        df   = pd.read_csv(dirr)
        fs   = np.array(df['frequency'])
        fs   = fs*u.Hz
        MAs  = np.array(df['MA'])
        MTs  = np.array(df['MT'])
        DAEs = np.array(df['DAE'])

    else:
        dirr = dir0 + interf + '_response_f_X.csv'
        df   = pd.read_csv(dirr)
        fs   = np.array(df['frequency'])
        fs   = fs*u.Hz
        MAs  = np.array(df['MX'])
        MTs  = np.array(df['MXY'])
        DAEs = np.array(df['DXY'])

    # Note that for interferometry channels we have MA -> MX, MT -> MXY, DAE -> DXY

    return fs, MAs, MTs, DAEs

def read_sens(dir0=dir0, SNR=SNR_PLS, T=T_PLS, interf='LISA', Xi=False,
              TDI=True, chan='A'):

    """
    Function that reads the sensitivity Omega (expressed as a GW energy
    density spectrum), and the power law sensitivity (PLS)
    of the interferometer chosen (options are 'LISA', 'Taiji', and the
    LISA-Taiji network 'comb').

    For LISA and Taiji the sensitivity can be given for each chanel X, Y,
    Z or on the TDI chanels A=E, T (A chanel is the default option, and
    the relevant for Omega sensitivity).

    Arguments:
        dir0   -- directory where the sensitivity files are stored
                  (default 'detector_sensitivity')
        SNR    -- signal-to-noise ratio (SNR) of the resulting PLS (default 10)
        T      -- duration of the mission (in years) of the resulting PLS
                  (default 4)
        interf -- option to chose the interferometer (default 'LISA', other
                  options are 'Taiji', 'comb', i.e., LISA + Taiji, and
                  'muAres')
        Xi     -- option to return the helical sensitivity and PLS (default False),
                  available when 'LISA' or 'Taiji' interf are chosen
        TDI    -- option to compute sensitivity in TDI chanels A and T
                  (default True)
        chan   -- specific chanel (default 'A', other options are 'X', 'Y', 'Z' if no TDI,
                  and 'T' if TDI), only available for LISA and Taiji

    Returns:
        fs           -- array of frequencies
        interf_Om    -- sensitivity of the interferometer 'interf'
                        expressed as a GW energy density spectrum
                        (can be given for different chanels, see above)
        interf_OmPLS -- PLS of the interferometer 'interf'
        interf_Xi    -- helical sensitivity of the interferometer 'interf'
        interf_XiPLS -- helical PLS of the interferometer 'interf'
    """

    fact = SNR/np.sqrt(T)

    if interf == 'LISA':

        fs, LISA_Om    = read_csv('LISA_Omega',    dir0=dir0)
        fs, LISA_OmPLS = read_csv('LISA_OmegaPLS', dir0=dir0)
        LISA_OmPLS    *= fact
        if Xi:
            fs, LISA_Xi    = read_csv('LISA_Xi',    dir0=dir0, b='Xi')
            fs, LISA_XiPLS = read_csv('LISA_XiPLS', dir0=dir0, b='Xi')
            LISA_XiPLS *= fact
            return fs, LISA_Om, LISA_OmPLS, LISA_Xi, LISA_XiPLS
        else: return fs, LISA_Om, LISA_OmPLS

    if interf == 'Taiji':

        fs, Taiji_Om    = read_csv('Taiji_Omega',    dir0=dir0)
        fs, Taiji_OmPLS = read_csv('Taiji_OmegaPLS', dir0=dir0)
        Taiji_OmPLS    *= fact
        if Xi:
            fs, Taiji_Xi    = read_csv('Taiji_Xi',    dir0=dir0, b='Xi')
            fs, Taiji_XiPLS = read_csv('Taiji_XiPLS', dir0=dir0, b='Xi')
            Taiji_XiPLS    *= fact
            return fs, Taiji_Om, Taiji_OmPLS, Taiji_Xi, Taiji_XiPLS
        else: return fs, Taiji_Om, Taiji_OmPLS

    if interf == 'comb':

        fs, LISA_Taiji_Xi    = read_csv('LISA_Taiji_Xi',    dir0=dir0, b='Xi')
        fs, LISA_Taiji_XiPLS = read_csv('LISA_Taiji_XiPLS', dir0=dir0, b='Xi')
        LISA_Taiji_XiPLS    *= fact
        return fs, LISA_Taiji_Xi, LISA_Taiji_XiPLS

    if interf == 'muAres':

        fs, muAres_Om    = read_csv('muAres_Omega',    dir0=dir0)
        fs, muAres_OmPLS = read_csv('muAres_OmegaPLS', dir0=dir0)
        muAres_OmPLS    *= fact
        return fs, muAres_Om, muAres_OmPLS

def read_csv(file, dir0=dir0, a='f', b='Omega'):

    """
    Function that reads a csv file with two arrays and returns them.

    Arguments:
        dir0 -- directory that contains the file (default is 'detector_sensitivity')
        file -- name of the csv file
        a -- identifier in pandas dataframe of first array (default 'f')
        b -- identifier in pandas dataframe of second array (default 'Omega')

    Returns:
        x -- first array of the file
        y -- second array of the file
    """

    dirr = dir0 + file + '.csv'
    df   = pd.read_csv(dirr)
    x    = np.array(df[a])
    y    = np.array(df[b])

    return x, y

def read_detector_PLIS_Schmitz(dir0=dir0 + '/power-law-integrated_sensitivities/',
                               det='BBO', SNR=SNR_PLS, T=T_PLS):

    """
    Read power law integrated senstivities from reference Schmitz:2020syl.

    Arguments:
        dir0 -- directory where the PLS are stored (default is
                'detector_sensitivity/power-law-integrated_sensitivities')
        det  -- GW detector (check available detectors in default directory)
        SNR  -- signal-to-noise ratio (SNR) of the resulting PLS (default 10)
        T    -- duration of the mission (in years) of the resulting PLS
                (default 4)
    """

    frac  = SNR/np.sqrt(T)
    dirr  = dir0 + 'plis_' + det + '.dat'
    BBO   = pd.read_csv(dirrr, header=14, delimiter='\t',
                        names=['f', 'Omega (log)', 'hc (log)', 'Sh (log)'])
    f     = 10**np.array(BBO['f'])
    Omega = 10**np.array(BBO['Omega (log)'])

    return f, Omega*frac

def read_MAC(dir0=dir0 + '/LISA_Taiji/', M='MAC', V='V'):

    """
    Function that reads the V response functions of the cross-correlated
    channels of the LISA-Taiji network.

    Reference is RoperPol:2021xnd, see figure 18. The data is taken from  Orlando:2020oko, see figure 2.

    Argument:
        dir0 -- directory where to save the results (default is
                'detector_sensitivity/LISA_Taiji/')
        M -- string of the channels to be read (options are default 'MAC',
             'MAD', 'MEC', 'MED')
        V -- Stokes parameter. It can be changed to read the 'I' response
             function (default 'V')
    """

    dirr = dir0 + M + '_' + V + '.csv'
    df   = pd.read_csv(dirr)
    f    = np.array(df['f'])
    MAC  = np.array(df['M'])
    inds = np.argsort(f)
    f    = f[inds]
    MAC  = MAC[inds]
    f    = f*u.Hz

    return f, MAC

def read_all_MAC(V='V'):

    """
    Function that reads all relevant TDI cross-correlated response functions between LISA
    and Taiji (AC, AD, EC, ED channels) using read_MAC.

    Reference is RoperPol:2021xnd, see figure 18. The data is taken from  Orlando:2020oko, see figure 2.

    Argument:
         V -- Stokes parameter. It can be changed to read the 'I' response
             function (default 'V')
    """

    f_AC, M_AC = read_MAC(M='MAC', V=V)
    f_AD, M_AD = read_MAC(M='MAD', V=V)
    f_EC, M_EC = read_MAC(M='MEC', V=V)
    f_ED, M_ED = read_MAC(M='MED', V=V)

    min_f = np.max([np.min(f_AC.value), np.min(f_AD.value),
                    np.min(f_EC.value), np.min(f_ED.value)])
    max_f = np.min([np.max(f_AC.value), np.max(f_AD.value),
                    np.max(f_EC.value), np.max(f_ED.value)])

    fs   = np.logspace(np.log10(min_f), np.log10(max_f), 1000)*u.Hz
    M_AC = np.interp(fs, f_AC, M_AC)*2
    M_AD = np.interp(fs, f_AD, M_AD)*2
    M_EC = np.interp(fs, f_EC, M_EC)*2
    M_ED = np.interp(fs, f_ED, M_ED)*2

    return fs, M_AC, M_AD, M_EC, M_ED

####### NOISE POWER SPECTRAL DENSITY FUNCTIONS FOR SPACE-BASED INTERFEROMETERS

def Poms_f(f=f_ref, P=P_LISA, L=L_LISA):

    """
    Function that computes the power spectral density (PSD) of the optical
    metrology system (oms) noise.

    Reference is RoperPol:2021xnd, equation B.24.

    Arguments:
        f -- frequency array (should be in units of Hz)
        P -- noise parameter (default 15 for LISA)
             value for Taiji is P = 8
        L -- length of the interferometer arm (default is L = 2.5e6 km for LISA)

    Returns:
        Poms -- oms PSD noise
    """

    f_mHz = f.to(u.mHz)
    L_pm  = L.to(u.pm)
    Poms  = P**2/L_pm.value**2*(1 + (2/f_mHz.value)**4)/u.Hz

    return Poms

def Pacc_f(f=f_ref, A=A_LISA, L=L_LISA):

    """
    Function that computes the power spectral density (PSD) of the mass
    acceleration noise.

    Reference is RoperPol:2021xnd, equation B.25.

    Arguments:
        f -- frequency array (should be in units of Hz)
        A -- noise parameter (default 3 for LISA)
             value for Taiji is A = 3
        L -- length of the interferometer arm (default is L = 2.5e6 km for LISA)

    Returns:
        Pacc -- mass acceleration PSD noise
    """

    f_mHz = f.to(u.mHz)
    L_fm  = L.to(u.fm)
    c     = const.c
    Loc   = L/c
    Loc   = Loc.to(u.s)

    fsinv = (c/2/np.pi/f/L)
    fsinv = fsinv.to(1)

    Pacc  = A**2*Loc.value**4/L_fm.value**2*(1 + (.4/f_mHz.value)**2)*(1 + \
            (f_mHz.value/8)**4)*fsinv.value**4/u.Hz

    return Pacc

def Pn_f(f=f_ref, P=P_LISA, A=A_LISA, L=L_LISA, TDI=True):

    """
    Function that computes the noise power spectral density (PSD) of an
    interferometer channel X, Pn(f), and of the cross-correlation of two
    different interferometer channels XY, Pn_cross(f). It gives the A and
    T PSD of the TDI chanels if TDI is True (default)

    Reference is RoperPol:2021xnd, equations B.23 and B.26.

    Arguments:
        f -- frequency array (should be in units of Hz)
        P -- noise parameter (default 15 for LISA)
        A -- noise parameter (default 3 for LISA)
        L -- length of the interferometer arm (default is L = 2.5e6 km for LISA,
             otherwise should be in units of km)

    Returns:
        Pn (noise PSD) and Pn_cross (cross-correlation noise PSD) if TDI is False,
            or PnA and PnT (noise PSD of the TDI channels A and T) if TDI is True
    """

    Poms = Poms_f(f=f, P=P, L=L)
    Pacc = Pacc_f(f=f, A=A, L=L)
    c    = const.c
    f0   = c/2/np.pi/L
    f_f0 = f.to(u.Hz)/f0.to(u.Hz)

    Pn       = Poms + (3 + np.cos(2*f_f0.value))*Pacc
    Pn_cross = -.5*np.cos(f_f0.value)*(Poms + 4*Pacc)

    if TDI:
        PnA = 2*(Pn - Pn_cross)/3
        PnT = (Pn + 2*Pn_cross)/3
        return PnA, PnT

    else:
        return Pn, Pn_cross

############################ INTERFEROMETRY CALCULATIONS ############################

##### ANALYTICAL FIT FOR LISA SENSITIVITY

def R_f(f=f_ref, L=L_LISA):

    """
    Function that computes the analytical fit of the response function.

    Reference is RoperPol:2021xnd, equation B.15.

    Arguments:
        f -- frequency array (should be in units of Hz)
        L -- length of the interferometer arm (default is L = 2.5e6 km for LISA,
             otherwise should be in units of km)

    Returns:
        Rf -- response function (using analytical fit)
    """

    c    = const.c
    f0   = c/2/np.pi/L
    f_f0 = f.to(u.Hz)/f0.to(u.Hz)
    Rf   = .3/(1 + .6*f_f0.value**2)

    return Rf

def Sn_f_analytical(f=f_ref, P=P_LISA, A=A_LISA, L=L_LISA):

    """
    Function that computes the strain sensitivity using the analytical fit
    for an interferometer channel X.

    Arguments:
        f  -- frequency array (should be in units of Hz)
        P  -- noise parameter (default 15 for LISA)
        A  -- noise parameter (default 3 for LISA)
        L  -- length of the interferometer arm (default is L = 2.5e6 km for LISA,
              otherwise should be in units of km)

    Returns:
        Rf -- response function (using analytical fit)
    """

    Pn = Pn_f(f=f, P=P, A=A, L=L)
    Rf = R_f(f=f,  L=L)

    return Pn/Rf

######################## NUMERICAL COMPUTATION OF RESPONSE FUNCTIONS ########################

def delta(a, b):

    """
    Function that returns the Kronecker delta delta(a, b).
    Returns 1 if a = b and 0 otherwise.
    """

    if a==b: return 1
    else:    return 0

def compute_interferometry(f=f_ref, L=L_LISA, TDI=True, order=1, comp_all=False,
                           comp_all_rel=True):

    """
    Function that computes numerically the monopole (or dipole if order = 2)
    response functions of an interferometer channel of a space-based GW detector
    using the interferometer channels (or TDI channels if TDI is True).

    Reference is RoperPol:2021xnd, appendix B (in particular, equations B.13 and B.16).

    Arguments:
        f            -- frequency array (should be in units of Hz)
        L            -- length of the interferometer arm
                        (default is L = 2.5e6 km for LISA, otherwise should be
                        in units of km)
        TDI          -- option to compute sensitivity in TDI chanels A and T
                        (default True), note that dipole is only computed for TDI.
        order        -- moment of the response function (default 1 corresponds
                        to monopole response functions, order = 2 corresponds
                        to dipole response function).
        comp_all     -- computes all response functions (monopole and dipole,
                        X and A channels), allows to compute everything faster
                        than rerunning the code for each case (default False)
        comp_all_rel -- computes only (and all) relevant response functions
                        (i.e., the ones that are not identically zero or equal
                        to each other due to the geometric symmetries
                        of the interferometer)

    Returns:
        MXY -- if order = 1, returns monopole response functions where
                X, Y can be X, Y, Z (not TDI), A, E, T (TDI)
        DXY -- if order = 2, returns dipole response functions where X, Y
               can be X, Y, Z (not TDI), A, E, T (TDI)
    """

    if comp_all_rel: comp_all = True

    c = const.c

    # integration over sky directions (theta, phi)
    theta = np.linspace(0, np.pi, 50)
    phi   = np.linspace(0, 2*np.pi, 50)

    # array of wave numbers
    k  = 2*np.pi*f/c
    kL = L*k
    kL = kL.to(1)

    kLij, th, ph = np.meshgrid(kL, theta, phi, indexing='ij')

    kx1 = 0
    kx2 = np.cos(th)
    kx3 = .5*(np.sqrt(3)*np.cos(ph)*np.sin(th) + np.cos(th))

    kU1 = kx2 - kx1
    kU2 = kx3 - kx2
    kU3 = kx1 - kx3

    # detector transfer functions (eq. B.3)
    TkU1   = np.exp(-1j*kLij*(1+kU1)/2)*np.sinc(kLij.value*(1 - kU1)/2/np.pi)
    TkU1  += np.exp(1j*kLij*(1-kU1)/2)*np.sinc(kLij.value*(1 + kU1)/2/np.pi)
    TkU2   = np.exp(-1j*kLij*(1+kU2)/2)*np.sinc(kLij.value*(1 - kU2)/2/np.pi)
    TkU2  += np.exp(1j*kLij*(1-kU2)/2)*np.sinc(kLij.value*(1 + kU2)/2/np.pi)
    TkU3   = np.exp(-1j*kLij*(1+kU3)/2)*np.sinc(kLij.value*(1 - kU3)/2/np.pi)
    TkU3  += np.exp(1j*kLij*(1-kU3)/2)*np.sinc(kLij.value*(1 + kU3)/2/np.pi)
    TkmU1  = np.exp(-1j*kLij*(1-kU1)/2)*np.sinc(kLij.value*(1 + kU1)/2/np.pi)
    TkmU1 += np.exp(1j*kLij*(1+kU1)/2)*np.sinc(kLij.value*(1 - kU1)/2/np.pi)
    TkmU2  = np.exp(-1j*kLij*(1-kU2)/2)*np.sinc(kLij.value*(1 + kU2)/2/np.pi)
    TkmU2 += np.exp(1j*kLij*(1+kU2)/2)*np.sinc(kLij.value*(1 - kU2)/2/np.pi)
    TkmU3  = np.exp(-1j*kLij*(1-kU3)/2)*np.sinc(kLij.value*(1 + kU3)/2/np.pi)
    TkmU3 += np.exp(1j*kLij*(1+kU3)/2)*np.sinc(kLij.value*(1 - kU3)/2/np.pi)

    U1 = np.array([0, 0, 1])
    U2 = .5*np.array([np.sqrt(3), 0, -1])
    U3 = -.5*np.array([np.sqrt(3), 0, 1])

    c = np.matrix([[2, -1, -1], [0, -np.sqrt(3), np.sqrt(3)],
                  [1,1,1]])/3

    # interferometer response functions (eqs. B.1 and B.9)
    if TDI or comp_all:
        QA = np.zeros((3, 3, len(kL), len(theta), len(phi)))*1j
        QE = np.zeros((3, 3, len(kL), len(theta), len(phi)))*1j
        QT = np.zeros((3, 3, len(kL), len(theta), len(phi)))*1j
    if not TDI or comp_all:
        QX = np.zeros((3, 3, len(kL), len(theta), len(phi)))*1j
        QY = np.zeros((3, 3, len(kL), len(theta), len(phi)))*1j
        if not comp_all_rel:
            QZ = np.zeros((3, 3, len(kL), len(theta), len(phi)))*1j
    for i in range(0, 3):
        for j in range(0,3):
            Q1 = .25*np.exp(-1j*kLij*kx1)*(TkU1*U1[i]*U1[j] - TkmU3*U3[i]*U3[j])
            Q2 = .25*np.exp(-1j*kLij*kx2)*(TkU2*U2[i]*U2[j] - TkmU1*U1[i]*U1[j])
            Q3 = .25*np.exp(-1j*kLij*kx3)*(TkU3*U3[i]*U3[j] - TkmU2*U2[i]*U2[j])
            if TDI or comp_all:
                QA[i,j,:,:,:] = Q1*c[0,0] + Q2*c[0,1] + Q3*c[0,2]
                QE[i,j,:,:,:] = Q1*c[1,0] + Q2*c[1,1] + Q3*c[1,2]
                QT[i,j,:,:,:] = Q1*c[2,0] + Q2*c[2,1] + Q3*c[2,2]
            if not TDI or comp_all:
                QX[i,j,:,:,:] = Q1
                QY[i,j,:,:,:] = Q2
                if not comp_all_rel: QZ[i,j,:,:,:] = Q3

    k1 = np.cos(ph)*np.sin(th)
    k2 = np.sin(ph)*np.sin(th)
    k3 = np.cos(th)

    # polarization tensors (eq. B.14)
    e1ab = np.zeros((3, 3, len(kL), len(theta), len(phi)))*1j
    for i in range(0, 3):
        if i==0:   ki=k1
        elif i==1: ki=k2
        else:      ki=k3
        for j in range(0, 3):
            if j==0:   kj=k1
            elif j==1: kj=k2
            else:      kj=k3
            e1ab[i,j,:,:,:] = delta(i,j) - ki*kj
            if i==0:
                if j==1:   e1ab[i,j,:,:,:] += -1j*k3
                elif j==2: e1ab[i,j,:,:,:] += 1j*k2
            elif i==1:
                if j==0:   e1ab[i,j,:,:,:] += 1j*k3
                elif j==2: e1ab[i,j,:,:,:] += -1j*k1
            else:
                if j==0:   e1ab[i,j,:,:,:] += -1j*k2
                elif j==1: e1ab[i,j,:,:,:] += 1j*k1

    if TDI or comp_all:
        if order == 1 and not comp_all:
            print('computing TDI monopole response functions')
        if order == 2 and not comp_all:
            print('computing TDI dipole response functions')
        if comp_all:
            print('computing TDI monopole and dipole response functions')
        FAA = np.zeros((len(kL), len(theta), len(phi)))*1j
        FAE = np.zeros((len(kL), len(theta), len(phi)))*1j
        FTT = np.zeros((len(kL), len(theta), len(phi)))*1j
        if not comp_all_rel:
            FEE = np.zeros((len(kL), len(theta), len(phi)))*1j
            FAT = np.zeros((len(kL), len(theta), len(phi)))*1j
            FET = np.zeros((len(kL), len(theta), len(phi)))*1j
    if not TDI or comp_all:
        if order == 1 and not comp_all:
            print('computing interferometer monopole response functions')
        if order == 2 and not comp_all:
            print('computing interferometer dipole response functions')
        if comp_all:
            print('computing interferometer monopole and dipole response functions')
        FXX = np.zeros((len(kL), len(theta), len(phi)))*1j
        FXY = np.zeros((len(kL), len(theta), len(phi)))*1j
        if not comp_all_rel:
            FYY = np.zeros((len(kL), len(theta), len(phi)))*1j
            FZZ = np.zeros((len(kL), len(theta), len(phi)))*1j
            FXZ = np.zeros((len(kL), len(theta), len(phi)))*1j
            FYZ = np.zeros((len(kL), len(theta), len(phi)))*1j

    for a in range(0, 3):
        for b in range(0, 3):
            for c in range(0, 3):
                for d in range(0, 3):
                    eabcd = .25*e1ab[a,c,:,:,:]*e1ab[b,d,:,:,:]
                    if TDI or comp_all:
                        FAA += eabcd*QA[a,b,:,:,:]*np.conjugate(QA[c,d,:,:,:])
                        FAE += eabcd*QA[a,b,:,:,:]*np.conjugate(QE[c,d,:,:,:])
                        FTT += eabcd*QT[a,b,:,:,:]*np.conjugate(QT[c,d,:,:,:])
                        if not comp_all_rel:
                            FEE += eabcd*QE[a,b,:,:,:]*np.conjugate(QE[c,d,:,:,:])
                            FAT += eabcd*QA[a,b,:,:,:]*np.conjugate(QT[c,d,:,:,:])
                            FET += eabcd*QE[a,b,:,:,:]*np.conjugate(QT[c,d,:,:,:])
                    if not TDI or comp_all:
                        FXX += eabcd*QX[a,b,:,:,:]*np.conjugate(QX[c,d,:,:,:])
                        FXY += eabcd*QX[a,b,:,:,:]*np.conjugate(QY[c,d,:,:,:])
                        if not comp_all_rel:
                            FYY += eabcd*QY[a,b,:,:,:]*np.conjugate(QY[c,d,:,:,:])
                            FZZ += eabcd*QZ[a,b,:,:,:]*np.conjugate(QZ[c,d,:,:,:])
                            FXZ += eabcd*QX[a,b,:,:,:]*np.conjugate(QZ[c,d,:,:,:])
                            FYZ += eabcd*QY[a,b,:,:,:]*np.conjugate(QZ[c,d,:,:,:])

    # Monopole (eq. B.13) and dipole (eq. B.16) response functions of LISA for TDI channels
    if TDI or comp_all:

        if order == 1 or comp_all:
            MAA1 = np.trapz(FAA*np.sin(th), th, axis=1)
            MAA  = np.trapz(MAA1, phi, axis=1)/np.pi
            MTT1 = np.trapz(FTT*np.sin(th), th, axis=1)
            MTT  = np.trapz(MTT1, phi, axis=1)/np.pi
            if not comp_all_rel:
                MEE1 = np.trapz(FEE*np.sin(th), th, axis=1)
                MEE  = np.trapz(MEE1, phi, axis=1)/np.pi
                MAE1 = np.trapz(FAE*np.sin(th), th, axis=1)
                MAE  = np.trapz(MAE1, phi, axis=1)/np.pi
                MAT1 = np.trapz(FAT*np.sin(th), th, axis=1)
                MAT  = np.trapz(MAT1, phi, axis=1)/np.pi
                MET1 = np.trapz(FET*np.sin(th), th, axis=1)
                MET  = np.trapz(MET1, phi, axis=1)/np.pi

            if not comp_all:
                return MAA, MEE, MTT, MAE, MAT, MET

        if order == 2 or comp_all:
            DAE1 = 1j*np.trapz(FAE*np.sin(th)**2*np.sin(ph), th, axis=1)
            DAE = np.trapz(DAE1, phi, axis=1)/np.pi
            if not comp_all_rel:
                DAA1 = 1j*np.trapz(FAA*np.sin(th)**2*np.sin(ph), th, axis=1)
                DAA  = np.trapz(DAA1, phi, axis=1)/np.pi
                DEE1 = 1j*np.trapz(FEE*np.sin(th)**2*np.sin(ph), th, axis=1)
                DEE  = np.trapz(DEE1, phi, axis=1)/np.pi
                DTT1 = 1j*np.trapz(FTT*np.sin(th)**2*np.sin(ph), th, axis=1)
                DTT  = np.trapz(DTT1, phi, axis=1)/np.pi
                DAT1 = 1j*np.trapz(FAT*np.sin(th)**2*np.sin(ph), th, axis=1)
                DAT  = np.trapz(DAT1, phi, axis=1)/np.pi
                DET1 = 1j*np.trapz(FET*np.sin(th)**2*np.sin(ph), th, axis=1)
                DET  = np.trapz(DET1, phi, axis=1)/np.pi

            if not comp_all:
                return DAA, DEE, DTT, DAE, DAT, DET

    if not TDI or comp_all:

        if order == 1 or comp_all:
            MXX1 = np.trapz(FXX*np.sin(th), th, axis=1)
            MXX  = np.trapz(MXX1, phi, axis=1)/np.pi
            MXY1 = np.trapz(FXY*np.sin(th), th, axis=1)
            MXY  = np.trapz(MXY1, phi, axis=1)/np.pi
            if not comp_all_rel:
                MYY1 = np.trapz(FYY*np.sin(th), th, axis=1)
                MYY  = np.trapz(MYY1, phi, axis=1)/np.pi
                MZZ1 = np.trapz(FZZ*np.sin(th), th, axis=1)
                MZZ  = np.trapz(MZZ1, phi, axis=1)/np.pi
                MXZ1 = np.trapz(FXZ*np.sin(th), th, axis=1)
                MXZ  = np.trapz(MXZ1, phi, axis=1)/np.pi
                MYZ1 = np.trapz(FYZ*np.sin(th), th, axis=1)
                MYZ  = np.trapz(MYZ1, phi, axis=1)/np.pi

            if not comp_all:
                return MXX, MYY, MZZ, MXY, MXZ, MYZ

        if order == 2 or comp_all:
            DXY1 = 1j*np.trapz(FXY*np.sin(th)**2*np.sin(ph), th, axis=1)
            DXY = np.trapz(DXY1, phi, axis=1)/np.pi
            if not comp_all_rel:
                DXX1 = 1j*np.trapz(FXX*np.sin(th)**2*np.sin(ph), th, axis=1)
                DXX  = np.trapz(DXX1, phi, axis=1)/np.pi
                DYY1 = 1j*np.trapz(FYY*np.sin(th)**2*np.sin(ph), th, axis=1)
                DYY  = np.trapz(DYY1, phi, axis=1)/np.pi
                DZZ1 = 1j*np.trapz(FZZ*np.sin(th)**2*np.sin(ph), th, axis=1)
                DZZ  = np.trapz(DZZ1, phi, axis=1)/np.pi
                DXZ1 = 1j*np.trapz(FXZ*np.sin(th)**2*np.sin(ph), th, axis=1)
                DXZ  = np.trapz(DXZ1, phi, axis=1)/np.pi
                DYZ1 = 1j*np.trapz(FYZ*np.sin(th)**2*np.sin(ph), th, axis=1)
                DYZ  = np.trapz(DYZ1, phi, axis=1)/np.pi

            if not comp_all:
                return DXX, DYY, DZZ, DXY, DXZ, DYZ

    if comp_all:
        if comp_all_rel:
            return MAA, MTT, DAE, MXX, MXY, DXY
        else:
            return MAA, MEE, MTT, MAE, MAT, MET, DAA, DEE, DTT, DAE, DAT, DET, \
                   MXX, MYY, MZZ, MXY, MXZ, MYZ, DXX, DYY, DZZ, DXY, DXZ, DYZ

def refine_M(f, M, A=.3, exp=0):

    """
    Function that refines the response function by appending
    a A*f^exp in the low-frequency regime.

    Arguments:
        f   -- frequency array (should be in units of Hz)
        M   -- response function to be refined at low frequencies
        A   -- amplitude of the response function at low frequencies
               (default 0.3 for the LISA monopole response function)
        exp -- exponent of the response function at low frequencies (default 0
               for the LISA monopole response function)

    Returns:
        fs -- refined array of frequencies
        Ms -- refined response function
    """

    ff0 = np.logspace(-6, np.log10(f[0].value), 1000)*u.Hz
    fs  = np.append(ff0, f)
    Ms  = np.append(A*ff0.value**exp, np.real(M))
    return fs, Ms

def compute_response_LISA_Taiji(f=f_ref, dir0=dir0, save=True, ret=False):

    """
    Function that computes LISA and Taiji's monopole and dipole response functions
    using the 'compute_interferometry' routine. It only computes the relevant
    response functions (see the tutorial 'response_functions.ipynb' for details).

    Arguments:
        dir0 -- directory where to save the results (default is 'detector_sensitivity')
        save -- option to save the results as output files (default True) with name
                'LISA_response_f' and 'Taiji_response_f'
        ret  -- option to return the results from the function

    Returns:
        MAA -- monopole response function of the TDI channel A
        MTT -- monopole response function of the TDI channel T (Sagnac channel)
        DAE -- dipole response function of the TDI correlation of the
               channels A and E
        MXX -- monopole response function of the interferometer channel X
        MXY -- monopole response function of the correlation of interferometer
               channels X and Y
        DXY -- dipole response functions of the correlation of the interferometer
               channels X and Y
    """

    f = np.logspace(-4, 0, 5000)*u.Hz

    # LISA response functions
    print('Calculating LISA response functions')
    MAA, MTT, DAE, MXX, MXY, DXY = compute_interferometry(f=f, L=L_LISA, comp_all_rel=True)
    # Taiji response functions
    print('Calculating Taiji response functions')
    MAA_Tai, MTT_Tai, DAE_Tai, MXX_Tai, MXY_Tai, DXY_Tai = \
            compute_interferometry(f=f, L=L_Taiji, comp_all_rel=True)

    # refine response functions at low frequencies (from known results)
    fs, MAs      = refine_M(f, MAA)
    fs, MAs_Tai  = refine_M(f, MAA_Tai)
    fs, MTs      = refine_M(f, MTT, A=1.709840e6, exp=6)
    fs, MTs_Tai  = refine_M(f, MTT_Tai, A=5.105546e6, exp=6)
    fs, DAEs     = refine_M(f, DAE, A=.2)
    fs, DAEs_Tai = refine_M(f, DAE_Tai, A=.2)
    fs, MXs      = refine_M(f, MXX, A=MXX[0])
    fs, MXs_Tai  = refine_M(f, MXX_Tai, A=MXX_Tai[0])
    fs, MXYs     = refine_M(f, MXY, A=MXY[0])
    fs, MXYs_Tai = refine_M(f, MXY_Tai, A=MXY_Tai[0])
    fs, DXYs     = refine_M(f, DXY, A=DXY[0])
    fs, DXYs_Tai = refine_M(f, DXY_Tai, A=DXY_Tai[0])

    # Write response functions in csv files
    if save:
        df = pd.DataFrame({'frequency': fs, 'MX': np.real(MXs), 'MXY': np.real(MXYs),
                           'DXY': np.real(DXYs)})
        df.to_csv(dir0 + 'LISA_response_f_X.csv')
        df = pd.DataFrame({'frequency': fs, 'MA': MAs, 'MT': MTs,
                           'DAE': DAEs})
        df.to_csv(dir0 + 'LISA_response_f_TDI.csv')
        print('saved response functions of channels X, Y of LISA in ', dir0 + 'LISA_response_f_X.csv')
        print('saved response functions of TDI channels of LISA in ', dir0 + 'LISA_response_f_TDI.csv')
        df_Tai = pd.DataFrame({'frequency': fs, 'MX': np.real(MXs_Tai), 'MXY': np.real(MXYs_Tai),
                           'DXY': np.real(DXYs_Tai)})
        df_Tai.to_csv(dir0 + 'Taiji_response_f_X.csv')
        df_Tai = pd.DataFrame({'frequency': fs, 'MA': MAs_Tai, 'MT': MTs_Tai,
                           'DAE': DAEs_Tai})
        df_Tai.to_csv(dir0 + 'Taiji_response_f_TDI.csv')
        print('saved response functions of channels X, Y of Taiji in ', dir0 + 'Taiji_response_f_X.csv')
        print('saved response functions of TDI channels of Taiji in ', dir0 + 'Taiji_response_f_TDI.csv')

    if ret:
        return fs, MAs, MTs, DAEs, MXs, MXYs, DXYs, \
               MAs_Tai, MTs_Tai, DAEs_Tai, MXs_Tai, MXYs_Tai, DXYs_Tai

############################## SENSITIVITIES AND SNR ##############################

def Sn_f(fs=f_ref, interf='LISA', TDI=True, M='MED', Xi=False):

    """
    Function that computes the strain sensitivity using the analytical fit
    for an interferometer channel X.

    Arguments:
        interf -- option to chose the interferometer (default 'LISA', other
                  option availables are 'Taiji' and 'comb', referring to cross-correlated
                  channels of LISA and Taiji)
        TDI    -- option to read the response functions for TDI chanels, instead of XYZ
                  chanels (default True)
        M      -- selection of cross-correlated channels (only when interf='comb', default 'MED')
        V      -- selection of Stokes parameter (default intensity I)
        Xi     -- option to compute strain sensitivity to polarized GW backgrounds from the
                  anisotropies induced due to the Solar System proper motion (default False)

    Returns:
        fs  -- frequency array
        SnA -- strain sensitivity in the channel A (if TDI), X (if not TDI) or if Xi is True,
               in the cross-correlated channels A and E (if TDI) or X and Y (if not TDI)
        SnT -- strain sensitivity in the Sagnac channel T (if TDI), or in the cross-correlated
               X and Y channels (if not TDI), only returns SnT if Xi is False
    """

    # read LISA and Taiji TDI response functions
    if interf != 'comb':
        fs, MAs, MTs, DAEs = read_response_LISA_Taiji(TDI=TDI, interf=interf)
    else:
        if not Xi:
            f_ED_I, M_ED_I = read_MAC(M=M, V='I') # read correlated ED response
            if M == 'MED': f_ED_I, M_ED_I = refine_M(f_ED_I, M_ED_I, A=0.028277196782809974)
            M_ED_I *= 2 # to get monopole response function
            fs = np.logspace(np.log10(f_ED_I[0].value), np.log10(f_ED_I[-1].value), 1000)*u.Hz
            MAs = np.interp(fs, f_ED_I, M_ED_I)
        else:
            fs, M_AC, M_AD, M_EC, M_ED = read_all_MAC(V='V')
            if M == 'MAC': MAs = abs(M_AC)
            if M == 'MAD': MAs = abs(M_AD)
            if M == 'MEC': MAs = abs(M_EC)
            if M == 'MED': MAs = abs(M_ED)
        MTs = MAs**0

    ## power spectral density of the noise
    if interf == 'LISA':
        PnA, PnT = Pn_f(f=fs, TDI=TDI, P=P_LISA, A=A_LISA, L=L_LISA)
    if interf == 'Taiji':
        PnA, PnT = Pn_f(f=fs, TDI=TDI, P=P_Taiji, A=A_Taiji, L=L_Taiji)
    if interf == 'comb':
        PnA, PnT = Pn_f(f=fs, TDI=TDI, P=P_LISA, A=A_LISA, L=L_LISA)
        PnC, PnS = Pn_f(f=fs, TDI=TDI, P=P_Taiji, A=A_Taiji, L=L_Taiji)
        PnA = np.sqrt(PnA*PnC)
        PnT = np.sqrt(PnT*PnS)

    # if Xi is True, it only returns the strain sensitivity to cross-correlated
    # channels A and E
    if Xi == True:
        if interf != 'comb': SnA = PnA/v/abs(DAEs)
        else:                SnA = PnA/MAs
        return fs, SnA

    # if not TDI, then PnA -> PnX, PnT -> Pncross, MAs -> MXs, MTs -> MXYs
    else: return fs, PnA/MAs, PnT/MTs

def Oms(f, S, h0=1., comb=False, S2=[], S3=[], S4=[], Xi=False):

    """
    Function that returns the sensitivity Sh(f) in terms of the GW energy density Om(f)

    Reference for Omega is RoperPol:2021xnd, equation B.18 (seems like it might have
    a typo, need to investigate this!). Final PLS sensitivites are again correct
    for a single chanel, since the factor of 2 in the SNR compensates for the
    1/2 factor here.

    Reference of Xi is RoperPol:2021xnd, equation B.21.

    Reference for combined sensitivities is RoperPol:2021xnd, equations B.37 (GW energy density,
    combining LISA and Taiji TDI channels A and C), and B.41 (polarization, combining
    4 cross-correlation between LISA and Taiji TDI channels AE, AD, CE, CD).

    Strain sensitivities, Omega sensitivities, and OmGW PLS agree with those of reference
    Caprini:2019pxz, see equation 2.14.

    Arguments:
        f    -- frequency array (should be in units of Hz)
        S    -- strain sensitivity function
        h0   -- parameterizes the uncertainties (Hubble tension) on the value
                of the Hubble rate (default 1)
        Xi   -- option to compute strain sensitivity to polarized GW backgrounds from the
                anisotropies induced due to the Solar System proper motion (default False)
        comb -- option to combine two sensitivities (e.g. LISA-Taiji sensitivities)
        S2   -- if comb is True, then second sensitivity to be combined
        S3, S4 -- if comb and Xi are both True, then it combines 4 sensitivites by cross-correlating
                  each of the 4 combinations of channels between LISA and Taiji (for example)

    Returns:
        Omega -- GW energy density sensitivity
    """

    H0 = co.H0_ref*h0           # reference value of H0 from cosmology (100 km/s/Mpc)
    A = 4*np.pi**2/3/H0**2
    if Xi: A /= 2
    Omega = S*A*f**3

    # if we are combining more than one channel for the sensitivity, we average them
    # through the harmonic mean
    if comb:
        Omega2 = S2*A*f**3
        Omega  = 1/np.sqrt(1/Omega**2 + 1/Omega2**2)
        if Xi:
            Omega3 = S3*A*f**3
            Omega4 = S4*A*f**3
            Omega  = 1/np.sqrt(1/Omega**2 + 1/Omega2**2 + \
                     1/Omega3**2 + 1/Omega4**2)

    return Omega

def compute_Oms_LISA_Taiji(interf='LISA', TDI=True, h0=1.):

    """
    Function that reads the response functions for LISA and/or Taiji, computes the strain
    sensitivities and from those, the sensitivity to the GW energy density spectrum Omega_s

    Arguments:
         interf -- option to chose the interferometer (default 'LISA', other
                   option availables are 'Taiji' and 'comb', referring to cross-correlated
                   channels of LISA and Taiji)
         TDI    -- option to read the response functions for TDI chanels, instead of XYZ
                   chanels (default True)
    """

    # read LISA and Taiji strain sensitivities Sn_f(f)
    fs, SnA, SnT = Sn_f(interf=interf, TDI=TDI)

    # Sn is the sensitivity of the channel A (if TDI) or X (if not TDI) for LISA or
    # Taiji (depending on what is interf)
    OmSA = Oms(fs, SnA, h0=h0, comb=False, Xi=False)
    OmST = Oms(fs, SnT, h0=h0, comb=False, Xi=False)

    return fs, OmSA, OmST

def OmPLS(Oms, f=f_ref, beta=beta_ref, SNR=1, T=1, Xi=0):

    """
    Function that computes the power law sensitivity (PLS).
    Reference is RoperPol:2021xnd, appendix B (equation B.31).

    Arguments:
        Oms  -- GW energy density sensitivity
        f    -- frequency array (should be in units of Hz, default is f_ref)
        beta -- array of slopes (default is beta_ref from -20 to 20)
        SNR  -- threshold signal-to-noise ratio (SNR) to compute the PLS
                (default 1)
        T    -- duration of the observation (in units of year, default 1)
        Xi   -- allows to compute PLS for polarization signals using the dipole
                response function setting Xi = 0.25 (default 0)

    Returns:
        Omega -- GW energy density power law sensitivity (PLS)
    """

    Cbeta = np.zeros(len(beta))
    T     = (T*u.yr).to(u.s)
    for i in range(0, len(beta)):
        aux = f.value**(2*beta[i])
        aa  = abs(1 - Xi*beta[i])
        Cbeta[i] = SNR/aa/np.sqrt(np.trapz(aux/Oms**2, f.value)*T.value)

    funcs = np.zeros((len(f), len(beta)))
    for i in range(0, len(beta)): funcs[:,i] = f.value**beta[i]*Cbeta[i]
    Omega = np.zeros(len(f))
    for j in range(0, len(f)): Omega[j] = np.max(funcs[j,:])

    return Omega

def SNR(f, OmGW, fs, Oms, T=1.):

    """
    Function that computes the signal-to-noise ratio (SNR) of a GW signal as

    SNR = sqrt(T int (OmGW/Oms)^2 df)

    Reference is RoperPol:2021xnd, appendix B (equation B.30).

    Arguments:
        f -- frequency array of the GW signal
        OmGW -- GW energy density spectrum of the GW signal
        fs -- frequency array of the GW detector sensitivity
        Oms -- GW energy density sensitivity of the GW detector
        T -- duration of observations in years (default 1)

    Returns:
        SNR -- SNR of the GW signal
    """

    T = T*u.yr
    T = T.to(u.s)
    T = T.value
    f = f.to(u.Hz)
    f = f.value
    OmGW = np.interp(fs, f, OmGW)
    OmGW[np.where(fs < f[0])] = 0
    OmGW[np.where(fs > f[-1])] = 0
    integ = np.trapz((OmGW/Oms)**2, fs)
    SNR   = np.sqrt(T*integ)

    return SNR



"""
plot_sets.py is a Python routine that includes some of the settings used
in the generation of plots for the cosmoGW project.

Adapted from the original plot_sets in GW_turbulence
(https://github.com/AlbertoRoper/GW_turbulence),
created January 2021

Currently part of the cosmoGW code:

https://github.com/cosmoGW/cosmoGW/
https://github.com/cosmoGW/cosmoGW/blob/development/src/cosmoGW/plot_sets.py

Author: Alberto Roper Pol
Created: 01/01/2021 (GW_turbulence)
Updated: 13/03/2025 (release cosmoGW 1.0: https://pypi.org/project/cosmoGW)
"""

import matplotlib.pyplot as plt

# Some common settings for the general plots

plt.rc('text', usetex=True)
plt.rc('font', family='serif')
plt.rc('font', size=20)
plt.rcParams.update({'axes.labelsize': 'x-large',
                     'axes.titlesize': 'x-large',
                     'xtick.labelsize': 'x-large',
                     'ytick.labelsize': 'x-large',
                     'legend.fontsize': 'x-large'})

def axes_lines(ax=[], both=True):

    """
    Function that includes some default settings for the lines in the loglog
    plots.
    """

    if ax == []: ax = plt.gca()
    ax.tick_params(axis='y', direction='in', length=12)
    ax.tick_params(axis='x', direction='in', length=12, pad=10)
    ax.tick_params(which='minor', direction='in', length=6)
    if both:
        ax.yaxis.set_ticks_position('both')
        ax.xaxis.set_ticks_position('both')

def save_fig(dirr='', name='fig', form='pdf', axes=True):

    """
    Function that saves the current plot as a PDF using the generic
    settings described in axes_lines.

    Arguments:
        - dirr: directory where to save the figure (default is current dir)
        - name: name of the figure (default is 'fig')
        - form: format of the figure file (default is 'pdf')
        - axes: option to call axes_lines for axes formatting
                (default is True)
    """

    import os

    if axes: axes_lines()
    plts = dirr + 'plots/'
    try:
        os.mkdir(plts)
    except:
        print('Not possible to create directory', plts)

    figg = name + '.' + form
    print('Saving figure in', plts, figg)
    plt.savefig(plts + figg, bbox_inches='tight')



"""
GW_templates.py is a Python routine that contains analytical and semi-analytical
templates of cosmological GW backgrounds, usually based on spectral fits,
either from GW models (see GW_models) or from numerical simulations

Currently part of the cosmoGW code:

https://github.com/cosmoGW/cosmoGW/
https://github.com/cosmoGW/cosmoGW/blob/main/src/cosmoGW/GW_templates.py

Author: Alberto Roper Pol
Created: 01/12/2022
Updated: 31/08/2024
Updated: 04/06/2025 (release cosmoGW 1.0: https://pypi.org/project/cosmoGW)

Main references are:

Espinosa:2010hh - J. R. Espinosa, T. Konstandin, J. M. No, G. Servant,
"Energy Budget of Cosmological First-order Phase Transitions,"
JCAP 06 (2010) 028, arXiv:1004.4187

Hindmarsh:2017gnf - M. Hindmarsh, S. J. Huber, K. Rummukainen,
D. J. Weir, "Shape of the acoustic gravitational wave
power spectrum from a first order phase transition,"
Phys.Rev.D 96 (2017) 10, 103520, Phys.Rev.D 101 (2020) 8,
089902 (erratum), arXiv:1704.05871

Hindmarsh:2019phv - M. Hindmarsh, M. Hijazi, "Gravitational waves from
first order cosmological phase transitions in the Sound Shell Model,"
JCAP 12 (2019) 062, arXiv:1909.10040

Caprini:2019egz   - [LISA CosWG], "Detecting gravitational
waves from cosmological phase transitions with LISA: an update,"
JCAP 03 (2020) 024, arXiv:1910.13125

Hindmarsh:2020hop - M. Hindmarsh, M. Lueben, J. Lumma,
M. Pauly, "Phase transitions in the early universe,"
SciPost Phys. Lect. Notes 24 (2021), 1, arXiv:2008.09136

Jinno:2022mie     - R. Jinno, T. Konstandin, H. Rubira, I. Stomberg,
"Higgsless simulations of cosmological phase transitions and
gravitational waves," JCAP 02, 011 (2023), arxiv:2209.04369

RoperPol:2022iel  - A. Roper Pol, C. Caprini, A. Neronov,
D. Semikoz, "The gravitational wave signal from primordial
magnetic fields in the Pulsar Timing Array frequency band,"
Phys. Rev. D 105, 123502 (2022), arXiv:2201.05630

RoperPol:2023bqa  - A. Roper Pol, A. Neronov, C. Caprini, T. Boyer,
D. Semikoz, "LISA and γ-ray telescopes as multi-messenger probes of a
first-order cosmological phase transition," arXiv:2307.10744 (2023)

EPTA:2023xxk      - EPTA and InPTA Collaborations, "The second data
release from the European Pulsar Timing Array - IV. Implications
for massive black holes, dark matter, and the early Universe,"
Astron. Astrophys. 685, A94 (2024), arxiv:2306.16227

RoperPol:2023dzg  - A. Roper Pol, S. Procacci, C. Caprini,
"Characterization of the gravitational wave spectrum from sound waves within
the sound shell model," Phys. Rev. D 109, 063531 (2024), arXiv:2308.12943.

Caprini:2024gyk   - A. Roper Pol, I. Stomberg, C. Caprini, R. Jinno,
T. Konstandin, H. Rubira, "Gravitational waves from first-order
phase transitions: from weak to strong," JHEP, arxiv:2409.03651

Caprini:2024hue   - E. Madge, C. Caprini, R. Jinno, M. Lewicki,
M. Merchand, G. Nardini, M. Pieroni, A. Roper Pol, V. Vaskonen,
"Gravitational waves from first-order phase transitions in LISA:
reconstruction pipeline and physics interpretation,"
JCAP 10, 020 (2024), arxiv:2403.03723

RoperPol:2025b    - A. Roper Pol, A. Midiri, M. Salomé, C. Caprini,
"Modeling the gravitational wave spectrum from slowly decaying sources in the
early Universe: constant-in-time and coherent-decay models," in preparation

RoperPol:2025a    - A. Roper Pol, S. Procacci, A. S. Midiri,
C. Caprini, "Irrotational fluid perturbations from first-order phase
transitions," in preparation
"""

import numpy  as np
import pandas as pd
import matplotlib.pyplot     as plt
import cosmoGW.cosmoGW       as cGW
import cosmoGW.cosmology     as co
import cosmoGW.GW_analytical as an
import cosmoGW.GW_models     as mod
import cosmoGW.hydro_bubbles as hb

HOME = cGW.HOME

# reference values
cs2_ref    = hb.cs2_ref     # speed of sound squared
Oms_ref    = mod.Oms_ref    # reference source amplitude
                            # (fraction to radiation energy density)
lf_ref     = mod.lf_ref     # reference length scale of the source
                            # (normalized by the Hubble radius)

# reference values and constants (EWPT)
Tref     = cGW.Tref       # EWPT (100 GeV)
gref     = cGW.gref       # EWPT (100 dof)
Neff_ref = co.Neff_ref    # reference Neff = 3

'''
Reference values for turbulence template
(based on RoperPol:2023bqa and RoperPol:2022iel)
Template used in Caprini:2024hue for LISA and in EPTA:2023xxk

Note that the values used for a_turb, b_turb, bPi_vort, fPi
are found assuming that the spectrum of the source is defined
such that

<v^2> ~ 2 E*    int zeta(K) dlnK,   as in RoperPol:2025b

which is a different normalization than that in previous papers,
where zeta is defined such that

<v^2> ~ 2 E* k* int zeta(K) dK,     in RoperPol:2022iel

Hence, the values are considered for the former zeta (this choice
yields different coefficients. However, the final result is not
affected by this choice. See RoperPol:2025b for details
'''

a_turb   = an.a_ref    # Batchelor spectrum k^5
b_turb   = an.b_ref    # Kolmogorov spectrum k^(-2/3)
alp_turb = 6/17        # von Karman smoothness parameter, see RoperPol:2023bqa
bPi_vort = 3           # spectral slope found in the anisotropic stresses
                       # in the K to infinity limit
alpPi    = 2.15        # smoothness parameter for the anisotropic
                       # stresses obtained for a von Karman spectrum
fPi      = 1.25        # break frequency of the anisotropic stresses
                       # obtained for a von Karman spectrum
N_turb   = mod.N_turb  # ratio between the effective time duration of
                       # the source and the eddy turnover time,
                       # based on the simulations of RoperPol:2022iel,
                       # used in RoperPol:2023bqa
tdecay_ref = 'eddy'    # decaying time used in the constant-in-time model
                       # (default is eddy turnover time)

### Reference values for sound waves templates

OmGW_sw_ref = 1e-2    # normalized amplitude, based on the simulations of
                      # Hindmarsh:2017gnf
a_sw_ref    = 3       # low frequency slope f^3 found for GWs in
                      # the HL simulations, see Jinno:2022mie and Caprini:2024gyk,
                      # and the sound-shell model in RoperPol:2023dzg
b_sw_ref    = 1       # intermediate frequency slope f found for GWs in
                      # the HL simulations, see Jinno:2022mie and Caprini:2024gyk
c_sw_ref    = 3       # high frequency slope f^(-3) found for GWs in the
                      # HL simulations, see Jinno:2022mie and Caprini:2024gyk
dt0_ref     = mod.dt0_ref  # reference value for dt0 used in the Higgsless fits

# first and second peak smoothness parameters
alp1_sw_ref = 1.5; alp2_sw_ref = 0.5  # used in RoperPol:2023bqa
alp1_ssm    = 4;   alp2_ssm    = 2.   # used in Hindmarsh:2019phv
alp1_HL     = 3.6; alp2_HL     = 2.4  # found in Caprini:2024gyk
alp1_LISA   = 2.;  alp2_LISA   = 4.   # used in Caprini:2024hue

####### GWB TEMPLATES FOR SOUND WAVES AND TURBULENCE #######

'''
Main reference: RoperPol:2023bqa

Turbulence template is based on the constant-in-time model
developed in RoperPol:2022iel, extended details in RoperPol:2025b

Sound waves template is based on the sound-shell model presented
in Hindmarsh:2019phv, revised and extended in RoperPol:2023dzg
and RoperPol:2025a

Double broken power law for sound waves uses the fits presented in
RoperPol:2023bqa, adapted from the numerical results of Jinno:2022mie
and Caprini:2024gyk

Templates have been used in RoperPol:2023bqa, Caprini:2024hue,
EPTA:2023xxk

The Higgsless template for decaying compressional sources
is developed and used in Caprini:2024gyk
'''

############################ SOUND WAVES ############################

################# values from Higgsless simulations #################

def data_warning(boxsize=20):

    print('You are using values that are interpolated from numerical',
          'data with L/vw = ', boxsize)
    print('Take into account that only alpha = 0.0046, 0.05',
          ' and 0.5 are found in simulations for vws from 0.32 to 0.8')
    print('Values out of this range should be taken with care')

def interpolate_HL_vals(df, vws, alphas, value='Omega_tilde_int_extrap',
                        boxsize=40, numerical=False, quiet=False):

    '''
    Function that uses the numerical results from Caprini:2024gyk to
    interpolate them to different values of wall velocity and alpha

    The data is stored under
      src/CosmoGW/resources/higgsless/parameters_fit_sims.csv
    '''

    mult_alpha = isinstance(alphas, (list, tuple, np.ndarray))

    columns    = df.box_size == boxsize
    df2        = df[columns]
    val_alphas = np.unique(df2['alpha'])
    val_vws    = np.unique(df2['v_wall'])
    Omegas     = np.zeros((len(val_vws), len(val_alphas))) - 1e30
    for i in range(0, len(val_alphas)):
        for j in range(0, len(val_vws)):
            Om = np.array(df2[value][(df2.v_wall == val_vws[j])* \
                          (df2.alpha == val_alphas[i])])
            if (len(Om) > 0):
                Omegas[j, i] = Om
                # for curly_K0 we interpolate kappa0 = curly_K0/alpha*(1 + alpha)
                if value == 'curly_K_0_512':
                    Omegas[j, i] *= (1 + val_alphas[i])/val_alphas[i]

    # interpolate for all values of vws for the 3 values of
    # alpha
    Omss = np.zeros((len(vws), len(val_alphas)))
    for i in range(0, len(val_alphas)):
        inds  = np.where(Omegas[:, i] > -1e30)[0]
        Omss[:, i] = np.interp(vws, val_vws[inds], Omegas[inds, i])
        inds2 = np.where(Omegas[:, i] == -1e30)[0]
        Omegas[inds2, i] = np.interp(val_vws[inds2], val_vws[inds],
                                     Omegas[inds, i])

    # interpolate for all values of alpha
    if mult_alpha:
        Omsss = np.zeros((len(vws), len(alphas)))
        for i in range(0, len(vws)):
            Omsss[i, :] = np.interp(np.log10(alphas), np.log10(val_alphas),
                                    Omss[i, :])

        # for curly_K0 we interpolate kappa0 = curly_K0/alpha*(1 + alpha)
        if value == 'curly_K_0_512':
            _, alpsij  = np.meshgrid(val_vws, val_alphas, indexing='ij')
            Omegas    *= alpsij/(1 + alpsij)
            _, alpsij  = np.meshgrid(vws, alphas, indexing='ij')
            Omsss     *= alpsij/(1 + alpsij)
    else:
        Omsss = np.zeros(len(vws))
        for i in range(0, len(vws)):
            Omsss[i] = np.interp(np.log10(alphas), np.log10(val_alphas),
                                 Omss[i, :])
        # for curly_K0 we interpolate kappa0 = curly_K0/alpha*(1 + alpha)
        if value == 'curly_K_0_512':
            Omsss  *= alphas/(1 + alphas)
            Omegas *= val_alphas/(1 + val_alphas)

    if not quiet:
        data_warning(boxsize=boxsize)
        if not numerical:
            print('To see numerical values call interpolate_HL_vals function setting',
                  ' numerical to True')

    # if numerical is chosen, also return numerical values
    if numerical: return Omsss, Omegas, val_alphas, val_vws
    else:         return Omsss

###################### TEMPLATE FOR SOUND WAVES ######################

def ampl_GWB_sw(model='fixed_value', OmGW_sw=OmGW_sw_ref, vws=[],
                alphas=[], numerical=False, bs_HL=20, quiet=False):

    '''
    Reference for sound waves is RoperPol:2023bqa, equation 3.

    Value of Omgwtilde = 1e-2 is based on Hindmarsh:2019phv, Hindmarsh:2017gnf
    and used for model = 'fixed_value'

    Values of Omgwtilde from the simulation results of Caprini:2024gyk
    are used for model = 'higgsless' by interpolating from the numerical
    data, which is available for alpha = 0.0046, 0.05 and 0.5 and vws in
    0.32 to 0.8
    '''

    mult_alp = isinstance(alphas, (list, tuple, np.ndarray))
    mult_vw  = isinstance(vws,    (list, tuple, np.ndarray))

    if not mult_alp: alphas = [alphas]
    if not mult_vw:  vws    = [vws]

    if   model == 'fixed_value': Omegas = np.full((len(vws), len(alphas)), OmGW_sw)

    elif model == 'higgsless':

        val_str  = 'Omega_tilde_int_extrap'

        try:
            if len(np.shape(vws)) == 1 or len(np.shape(alphas)) == 1:
                if len(vws) == 0 or len(alphas) == 0:
                    print('Provide values of vws and alphas to use Higgsless model',
                          ' in ampl_GWB_sw')
                    return 0

        except:
            tst = True

        # take values from higgsless dataset
        dirr = HOME + 'resources/higgsless/parameters_fit_sims.csv'
        df   = pd.read_csv(dirr)

        if numerical:
            Omegas, Omnum, val_alphas, val_vws = \
                        interpolate_HL_vals(df, vws, alphas,
                                            value=val_str, boxsize=bs_HL,
                                            numerical=numerical, quiet=quiet)

        else: Omegas = interpolate_HL_vals(df, vws, alphas,
                            value=val_str, boxsize=bs_HL, quiet=quiet)

    else:
        print('Choose an available model for ampl_GWB_sw for sound waves')
        print('Available models are fixed_value and higgsless')
        return 0

    if   not mult_alp and not mult_vw: Omegas = Omegas[0, 0]
    elif not mult_alp: Omegas = Omegas[:, 0]
    elif not mult_vw:  Omegas = Omegas[0, :]

    if numerical and model == 'higgsless': return Omegas, Omnum, val_alphas, val_vws
    else:                                  return Omegas

def pref_GWB_sw(Oms=Oms_ref, lf=lf_ref, alpha=0, model='sound_waves',
                Nshock=1., b=0., expansion=True, beta=100, cs2=cs2_ref):

    '''
    Dependence of the GW spectrum from sound waves on the mean
    size of the bubbles lf = R* H_* and the kinetic energy density

        Oms = vrms^2 = <w v^2>/<w>,

    related to K = kappa alpha/(1 + alpha) used in Caprini:2024gyk as

        Oms   = K/Gamma = kappa alpha/(1 + cs2),     where
        Gamma = <w>/<rho> = (1 + cs2)/(1 + alpha)

    is the adiabatic index Gamma ~ (1 + cs2)/(1 + alpha)

    Note that RoperPol:2023dzg uses OmK = .5 vrms^2 so an extra factor of
    2 appears in its relation to K.

    We consider different models:

    - Model 'sound_waves' uses Caprini:2024gyk eqs. 2.16 and 2.30
        It corresponds to the linear growth with the source duration,
        which is set to

           tdur = Nshock tshock = Nshock R*/sqrt(Oms),

        assuming sound waves do not decay and their UETC is stationary:

           OmGW ~ K^2 R* tdur = K^2 R* Nshock tshock

        When the Universe expansion is included, tdur is substituted
        by the suppresion factor Upsilon describing the growth
        with the source duration. For a radiation-dominated Universe
        (see RoperPol:2023dzg):

           Upsilon = tdur/(1 + tdur)

        This choice includes Universe expansion and assumes sourcing
        occurs during the radiation-dominated era

           OmGW ~ K^2 R* Upsilon(tdur) = K^2 R* tdur/(1 + tdur)

       When Nshock = 1, tdur = tshock = R*/sqrt(Oms) and it becomes equation 3
       of RoperPol:2023bqa, based on Hindmarsh:2020hop, equation 8.24

    - Model 'decaying' uses Caprini:2024gyk, eq. 5.6
        It includes the decay of the source assuming
        a locally stationary UETC. For Minkowski space-time:

           OmGW ~ K2int R*,       K2int     = int K^2 dt,

        while for a radiation-dominated expanding Universe:

            OmGW ~ K2int_exp R*,  K2int_exp = int K^2/t^2 dt.

        We will assume a power law decay K(t > t0) = K0 (dt/dt0)^(-b)
        based on the numerical findings of Caprini:2024gyk,
        see K2int function in GW_models module

    Arguments:
        Oms    -- kinetic energy density, defined as vrms^2, such that
                  Oms = K/Gamma = (1 + cs2) kappa alpha
        lf     -- mean-size of the bubbles, given as a fraction
                  of the Hubble radius, input must be one value
        alpha  -- ratio of vacuum to radiation energy densities
        model  -- chooses the model (sound_waves or decaying)
        Nshock -- describes the duration time in units of the
                  shock time
        b      -- power law decay in time of the kinetic energy
                  source (default 0 recovers stationary sound waves)
        expansion -- option to consider flat Minkowski space-time
                     (if expansion is False, default) or a radiation-
                     dominated expanding Universe
        beta   -- rate of nucleation of the phase transition, input must
                  be one value
        cs2    -- square of the speed of sound (default is 1/3 for
                  radiation domination)


    Returns:
        pref -- prefactor in the GW spectrum, consisting on
                K2int_exp x R*
    '''

    if len(np.shape(alpha)) == 0:
        if alpha == 0:
            print('you need to give the value of alpha as input',
                          'for a correct result in pref_GWB_sw')
            alpha = np.zeros_like(Oms)

    Gamma = (1. + cs2)/(1. + alpha)
    K     = Gamma*Oms

    pref  = K**2*lf
    tdur  = Nshock*lf/np.sqrt(Oms)

    if   model == 'sound_waves':
        if expansion: pref *= 1./(1. + 1./tdur)
        else:         pref *= tdur

    elif model == 'decay':

        K2int = mod.K2int(tdur, K0=K, b=b, dt0=dt0_ref, expansion=expansion, beta=beta)
        pref  = K2int*lf

    return pref

def Sf_shape_sw(s, model='sw_LISA', Dw=1., a_sw=a_sw_ref, b_sw=b_sw_ref, c_sw=c_sw_ref,
                alp1_sw=0, alp2_sw=0, strength='weak', interpolate_HL=False,
                bs_k1HL=40, bs_k2HL=20, vws=[], alphas=[], quiet=False,
                interpolate_HL_n3=False, corrRs=True, cs2=cs2_ref):

    """
    Function that computes the GW spectral shape generated by sound waves
    based on different templates.

    Arguments:
        s     -- normalized wave number, divided by the mean bubbles
                 size, s = f R*
        model -- model for the sound-wave template (options are 'sw_SSM',
                 'sw_HL', 'sw_LISA', 'sw_LISAold', and 'sw_HLnew')
        Dw    -- ratio between peak frequencies, determined by the shell thickness,
                 note that different models use slightly different conventions for Dw
        a_sw, b_sw, c_sw -- slopes for sound wave template (default is 3, 1, 3)
        alp1_sw, alp2_sw -- transition parameters for sound wave template
                            (default values for each model are listed above)
        strength -- phase transition strength, used to determine peak2 in sw_HLnew
                    (unless interpolate_HL is True)
        interpolate_HL   -- option to use numerical data from Caprini:2024gyk to
                            estimate slope c_sw and peak1 and peak2 positions
        bsk1_HL, bsk2_HL -- box size of Higgsless simulations used to estimate the
                            the peaks (default is 20 for k2 and 40 for k1)
        vws, alphas      -- array of wall velocities and alphas for which the
                            parameters are to be estimated
        quiet            -- option to output a warning about the interpolation
                            method and its range of validity

    Returns:
        S -- spectral shape of the GW spectrum (still to be normalized)
             as a function of s = f R*.
             If interpolate_HL is True, it returns an array of size
             (s, alphas, vws)
    """

    mult_Dw = isinstance(Dw, (list, tuple, np.ndarray))

    # in some models, Dw can be a 0, 1, or 2d input (vw, alpha) dependence
    Dw_2d = False
    if model == 'sw_LISA': Dw_2d = True
    if model == 'sw_HLnew':
        if strength == 'weak' and not interpolate_HL:
            Dw_2d = True

    if Dw_2d:

            # sound-shell thickness Dw needs to be used as an input
            NDw = len(np.shape(Dw))
            if NDw == 1: s, Dw = np.meshgrid(s, Dw, indexing='ij')
            if NDw == 2:
                s0  = np.zeros((len(s), np.shape(Dw)[0], np.shape(Dw)[1]))
                Dw0 = np.zeros((len(s), np.shape(Dw)[0], np.shape(Dw)[1]))
                for i in range(0, np.shape(Dw)[0]):
                    for j in range(0, np.shape(Dw)[1]):
                        s0[:,  i, j] = s
                        Dw0[:, i, j] = Dw[i, j]
                s   = s0
                Dw  = Dw0

    if model == 'sw_LISAold':

        # Reference for sound waves based on simulations of Hindmarsh:2017gnf
        # is Caprini:2019egz (equation 30) with only one peak

        # peak positions
        peak1 = 2*np.pi/10
        s     = peak1*s

        S = s**3*(7/(4 + 3*s**2))**(7/2)

    elif model == 'sw_SSM':

        # Reference for sound waves based on Sound Shell Model (sw_SSM) is
        # RoperPol:2023bqa, equation 6, based on the results presented in
        # Hindmarsh:2019phv, equation 5.7
        # Uses Dw = |vw - cs|/max(vw, cs)

        if not mult_Dw: Dw = [Dw]
        s, Dw = np.meshgrid(s, Dw, indexing='ij')

        # uses a different slope at large frequencies to adapt the result
        # at intermediate ones
        if c_sw == 3: c_sw = 4
        # takes a different slope at small frequencies
        if a_sw == 3: a_sw = 9

        # amplitude such that S = 1 at s = 1/Dw
        m = (9*Dw**4 + 1)/(Dw**4 + 1)
        A = Dw**9*(1 + Dw**(-4))**2*(5/(5 - m))**(5/2)

        # peak positions
        peak1 = 1.
        peak2 = np.sqrt((5 - m)/m)/Dw

        if alp1_sw == 0: alp1_sw=alp1_ssm
        if alp2_sw == 0: alp2_sw=alp2_ssm

        S = A*an.smoothed_double_bPL(s, peak1, peak2, A=1., a=a_sw, b=b_sw,
                                     c=c_sw, alp1=alp1_sw, alp2=alp2_sw,
                                     alpha2=True)

        if not mult_Dw: S = S[:, 0]

    elif model == 'sw_HL':

        # Reference for sound waves based on Higgsless (sw_HL) simulations is
        # RoperPol:2023bqa, equation 7, based on the results presented in
        # Jinno:2022mie
        # Uses Dw = |vw - cs|/max(vw, cs)

        if not mult_Dw: Dw = [Dw]
        s, Dw = np.meshgrid(s, Dw, indexing='ij')

        # amplitude such that S = 1 at s = 1/Dw
        A = 16*(1 + Dw**(-3))**(2/3)*Dw**3

        # peak positions
        peak1 = 1.
        peak2 = np.sqrt(3)/Dw

        if alp1_sw == 0: alp1_sw=alp1_sw_ref
        if alp2_sw == 0: alp2_sw=alp2_sw_ref

        S = A*an.smoothed_double_bPL(s, peak1, peak2, A=1., a=a_sw, b=b_sw,
                                     c=c_sw, alp1=alp1_sw, alp2=alp2_sw)

        if not mult_Dw: S = S[:, 0]

    elif model == 'sw_LISA':

        # Reference for sound waves based on Higgsless simulations is
        # Caprini:2024hue (equation 2.8), based on the results presented in
        # Jinno:2022mie, see updated results and discussion in Caprini:2024gyk.
        # Uses Dw = xi_shell/max(vw, cs)

        # smoothness parameters
        if alp1_sw == 0: alp1_sw = alp1_LISA
        if alp2_sw == 0: alp2_sw = alp2_LISA

        # peak positions
        peak1 = 0.2
        peak2 = 0.5/Dw

        S = an.smoothed_double_bPL(s, peak1, peak2, A=1., a=a_sw, b=b_sw,
                                   c=c_sw, alp1=alp1_sw, alp2=alp2_sw, alpha2=True)

    elif model == 'sw_HLnew':

        # Reference for sound waves based on updated HL results (sw_HLnew)
        # is Caprini:2024gyk
        # Uses Dw = xi_shell/max(vw, cs)

        # smoothness parameters
        if alp1_sw == 0: alp1_sw = alp1_HL
        if alp2_sw == 0: alp2_sw = alp2_HL

        if not interpolate_HL:

            # peak positions
            peak1 = 0.4
            if   strength == 'weak':   peak2 = 0.5/Dw
            elif strength == 'interm': peak2 = 1.
            else:                      peak2 = 0.5

            S = an.smoothed_double_bPL(s, peak1, peak2, A=1., a=a_sw, b=b_sw,
                                       c=c_sw, alp1=alp1_sw, alp2=alp2_sw, alpha2=True)

        else:

            if len(vws) == 0 or len(alphas) == 0:
                print('To use interpolate_HL in Sf_shape_sw',
                      ' give values of vws and alphas')
                return 0

            # take values from higgsless dataset
            dirr     = HOME + 'resources/higgsless/parameters_fit_sims.csv'
            df       = pd.read_csv(dirr)

            val_str  = 'k1'
            peaks1   = interpolate_HL_vals(df, vws, alphas, quiet=True,
                                   value=val_str, boxsize=bs_k1HL)
            val_str  = 'k2'
            peaks2   = interpolate_HL_vals(df, vws, alphas, quiet=True,
                                   value=val_str, boxsize=bs_k2HL)

            s0      = np.zeros((len(s), len(vws), len(alphas)))
            peaks10 = np.zeros((len(s), len(vws), len(alphas)))
            peaks20 = np.zeros((len(s), len(vws), len(alphas)))

            Rstar_beta = hb.Rstar_beta(vws=vws, corr=corrRs, cs2=cs2)/2/np.pi
            for i in range(0, len(vws)):
                for j in range(0, len(alphas)):
                    peaks10[:, i, j] = peaks1[i, j]*Rstar_beta[i]
                    peaks20[:, i, j] = peaks2[i, j]*Rstar_beta[i]
                    s0[:, i, j]      = s

            peaks1 = peaks10
            peaks2 = peaks20
            s      = s0

            if interpolate_HL_n3:
                val_str  = 'n3'
                c_sw     = - interpolate_HL_vals(df, vws, alphas, quiet=True,
                                       value=val_str, boxsize=bs_k2HL)

            if not quiet: data_warning(boxsize='%i and %i'%(bs_k1HL,bs_k2HL))

            S = an.smoothed_double_bPL(s, peaks1, peaks2, A=1., a=a_sw, b=b_sw,
                                       c=c_sw, alp1=alp1_sw, alp2=alp2_sw, alpha2=True)

    else:
        print('Choose an available model in Sf_shape_sw for the sound wave',
              'spectral shape.')
        print('Available models are sw_LISAold, sw_SSM, sw_HL, sw_LISA, sw_HLnew')
        return 0

    return S

def OmGW_spec_sw(s, alphas, betas, vws=1., cs2=cs2_ref, quiet=True, a_sw=a_sw_ref,
                 b_sw=b_sw_ref, c_sw=c_sw_ref, alp1_sw=0, alp2_sw=0, corrRs=True,
                 expansion=False, Nsh=1.,
                 model_efficiency='fixed_value', OmGW_tilde=OmGW_sw_ref,
                 bs_HL_eff=20, model_K0='Espinosa', bs_k1HL=40,
                 model_decay='sound_waves', interpolate_HL_decay=True, b=0,
                 model_shape='sw_LISA', strength='weak', interpolate_HL_shape=False,
                 interpolate_HL_n3=False, redshift=False, gstar=gref, gS=0, T=Tref,
                 h0=1., Neff=Neff_ref):

    '''
    Function that computes the GW spectrum (normalized to radiation
    energy density within RD era) for sound waves.

    The general shape of the GW spectrum is based on that of reference
    RoperPol:2023bqa, equations 3 and 9:

        OmGW (f) = 3 * ampl_GWB * pref_GWB * S(f),

    where:

    - ampl_GWB is the efficiency of GW production by the specific source,
    - pref_GWB is the dependence of the GW amplitude on the source
      parameters (e.g. length scale and strength of the source),
    - S(f) is a normalized spectral shape, such that int S(f) d ln f = 1

    The GW spectrum as an observable at present time is then computed using

        OmGW0 (f) = OmGW x FGW0,

    where FGW0 is the redshift from the time of generation to present
    time, computed in cosmoGW.py that depends on the degrees of freedom at
    the time of generation.

    Arguments:
        s                    -- normalized wave number, divided by the mean bubbles
                                size Rstar, s = f R*
        alphas               -- strength of the phase transition
        betas                -- rate of nucleation of the phase transition
        vws                  -- array of wall velocities
        cs2                  -- square of the speed of sound
                                (default is 1/3 for radiation domination)
        quiet                -- option to avoid printing debugging info (default is True)
        a_sw, b_sw, c_sw     -- slopes of the sound wave template
                                (takes 3, 1, 3 by default)
        alp1_sw, alp2_sw     -- transition parameters of the sound wave template
                                (takes the template values by default of each model_shape)
        corr_Rs              -- option to correct the ratio R* beta with max(vw, cs)
        expansion            -- option to include the Universe expansion
                                (in radiation domination)
        Nsh                  -- number of shock formation times to determine the
                                source duration (default is 1)
        model_efficiency     -- model to compute Omega tilde (default is 'fixed_value',
                                other option available is 'higgsless')
        OmGW_tilde           -- value of Omega tilde used when
                                model_efficiency='fixed_value' (default is 1e-2)
        bs_HL_eff            -- box size of the Higgsless simulations used to interpolate
                                the values of Omega tilde when model_efficiency='higgsless'
                                (default if L/vw = 20)
        model_K0             -- model to compute the kinetic energy ratio K0
                                (default is 'Espinosa' to use the the single-bubble value,
                                other option is 'higgsless')
        bs_k1HL              -- box size of the Higgsless simulations used to interpolate
                                the values of k1 when model_shape='higgsless'
                                (default if L/vw = 20)
        model_decay          -- model to compute the prefactor of the GW amplitude
                                (default is 'sound_waves' to consider stationary sourcing,
                                other option is 'decay')
        interpolate_HL_decay -- option to use Higgsless simulations to interpolate the
                                values of b (default is False)
        b                    -- decay law in time of the kinetic energy density
                                (default is 0)
        model_shape          -- model used to compute the spectral shape (options
                                are 'sw_LISAold', 'sw_SSM', 'sw_HL', 'sw_LISA',
                                'sw_HLnew')
        strength             -- strength of the phase transition to use average
                                values of k1 and k2 when model_shape = 'sw_HLnew'
                                is used (default is 'weak', other options are
                                'interm' and 'strong')
        interpolate_HL_shape -- option to use Higgsless simulations to interpolate the
                                values of k1 and k2 (default is False)
        interpolate_HL_n3    -- option to use Higgsless simulations to interpolate the
                                values of c_sw (default is False)
        redshift             -- option to redshift the GW spectrum and frequencies to
                                present time (default is False)
        gstar                -- number of degrees of freedom (default is 100) when
                                redshift = True
        gS                   -- number of adiabatic degrees of freedom (default is gstar)
                                when redshift = True
        T                    -- temperature scale (default is 100 GeV) when redshift = True
        h0                   -- value of the Hubble rate at present time (100 h0 Mpc/km/s)
                                (default is one)
        Neff                 -- effective number of neutrino species (default is 3)

    Returns:
        freq                 -- array of frequencies (normalized with Rstar if redshift
                                is False and in Hz otherwise)
        OmGW                 -- GW spectrum normalized to the radiation energy density
                                (or to the present critical density if redshift is True)
    '''

    cs         = np.sqrt(cs2)
    mult_alpha = isinstance(alphas, (list, tuple, np.ndarray))
    mult_beta  = isinstance(betas,  (list, tuple, np.ndarray))
    mult_vws   = isinstance(vws,    (list, tuple, np.ndarray))

    if not mult_alpha: alphas = np.array([alphas])
    if not mult_vws:   vws    = np.array([vws])
    if not mult_beta:  betas  = np.array([betas])

    #### Computing ampl_GWB

    if model_efficiency == 'higgsless' and not quiet:
        print('Computing the OmGW efficiency')
        data_warning(boxsize=bs_HL_eff)

    ampl = ampl_GWB_sw(model=model_efficiency, OmGW_sw=OmGW_tilde,
                       vws=vws, alphas=alphas, bs_HL=bs_HL_eff, quiet=True)

    #### Computing pref_GWB

    # Kinetic energy density

    if not quiet:
        print('Computing the kinetic energy density using the model',
              model_K0)
        if model_K0 == 'higgsless': data_warning(boxsize=bs_HL_eff)

    if model_K0 == 'Espinosa':

        # compute kappa, K and Oms following the bag equation of
        # state as in Espinosa:2010hh

        # kappa: efficiency in converting vacuum to kinetic energy
        # K = rho_kin/rho_total = kappa alpha/(1 + alpha)
        # Oms_sw = v_f^2 = kappa alpha/(1 + cs2)
        kap    = hb.kappas_Esp(vws, alphas, cs2=cs2)
        K      = kap*alphas/(1 + alphas)
        Oms_sw = kap*alphas/(1 + cs2)

    elif model_K0 == 'higgsless':

        # compute K and Oms directly from the numerical results of the
        # Higgsless simulations of Caprini:2024gyk and interpolate to
        # values of alpha and vws

        dirr    = HOME + 'resources/higgsless/parameters_fit_sims.csv'
        df      = pd.read_csv(dirr)
        val_str = 'curly_K_0_512'
        K       = interpolate_HL_vals(df, vws, alphas, quiet=quiet,
                                      value=val_str, boxsize=bs_HL_eff)
        kap     = K*(1 + alphas)/alphas
        Oms_sw  = kap*alphas/(1 + cs2)

    else:
        print('Choose an available model for K0 in OmGW_spec_sw')
        print('Available models are Espinosa and higgsless')
        return 0

    # Decay rate

    interpol_b = False
    if interpolate_HL_decay and model_decay == 'decay':

        dirr    = HOME + 'resources/higgsless/parameters_fit_sims.csv'
        df      = pd.read_csv(dirr)

        val_str = 'b'
        b       = interpolate_HL_vals(df, vws, alphas, quiet=quiet,
                                          value=val_str, boxsize=bs_HL_eff)
        interpol_b = True

    # prefactor GWB of sound waves

    pref = np.zeros((len(vws), len(alphas), len(betas)))
    for i in range(0, len(vws)):
        # Fluid length scale R_star x beta
        lf = hb.Rstar_beta(vws[i], cs2=cs2, corr=corrRs)/betas
        for j in range(0, len(alphas)):
            for l in range(0, len(betas)):
                if interpol_b: b_ij = b[i, j]
                else:          b_ij = b
                pref[i, j, l] = pref_GWB_sw(Oms=Oms_sw[i, j], lf=lf[l], alpha=alphas[j],
                                    model=model_decay, Nshock=Nsh, b=b_ij,
                                    expansion=expansion, beta=betas[l], cs2=cs2)

    #### Computing the spectral shape

    # sound-shell thickness and spectral shape

    if not quiet: print('Computing spectral shape using model ',
                        model_shape)

    if model_shape == ['sw_LISAold']:

        S  = Sf_shape_sw(s,  model=model_shape)
        try:    mu = np.trapezoid(S, np.log(s))
        except: mu = np.trapz(S, np.log(s))
        # normalized spectral shape
        S  = S/mu
        S, _, _ = np.meshgrid(S, vws, alphas, indexing='ij')

    elif model_shape in ['sw_HL', 'sw_SSM']:

        Dw = abs(vws - cs)/vws
        S  = Sf_shape_sw(s, model=model_shape, Dw=Dw, a_sw=a_sw, b_sw=b_sw, c_sw=c_sw,
                         alp1_sw=alp1_sw, alp2_sw=alp2_sw)
        try:    mu = np.trapezoid(S, np.log(s), axis=0)
        except: mu = np.trapz(S, np.log(s), axis=0)
        S  = S/mu
        S0 = np.zeros((len(s), len(vws), len(alphas)))
        for i in range(0, len(alphas)):
            S0[:, :, i] = S
        S  = S0

    elif model_shape in ['sw_LISA', 'sw_HLnew']:

        # in some models, Dw is required to be computed for the array
        # of vws and alphas
        Dw_2d = True
        if model_shape == 'sw_HLnew':
            if strength != 'weak' or interpolate_HL_shape: Dw_2d = False

        if Dw_2d:
            if not quiet: print('Computing sound-shell thickness')
            _, _, _, _, _, _, xi_shocks, _ = \
                            hb.compute_profiles_vws_multalp(alphas, vws=vws)

            Dw = np.zeros((len(vws), len(alphas)))
            for i in range(0, len(alphas)):
                Dw[:, i] = (xi_shocks[:, i] - np.minimum(vws, cs))/np.maximum(vws, cs)

        else: Dw = 0.

        S = Sf_shape_sw(s, model=model_shape, Dw=Dw, a_sw=a_sw, b_sw=b_sw, c_sw=c_sw,
                alp1_sw=alp1_sw, alp2_sw=alp2_sw, strength=strength,
                interpolate_HL=interpolate_HL_shape,
                bs_k1HL=bs_k1HL, bs_k2HL=bs_HL_eff, vws=vws, alphas=alphas, quiet=quiet,
                interpolate_HL_n3=interpolate_HL_n3, corrRs=corrRs, cs2=cs2)

        try:    mu = np.trapezoid(S, np.log(s), axis=0)
        except: mu = np.trapz(S, np.log(s), axis=0)
        S  = S/mu

        if not interpolate_HL_shape:
            if strength != 'weak':
                S, _, _ = np.meshgrid(S, vws, alphas, indexing='ij')

    else:
        print('Choose an available model for model_shape in OmGW_spec_sw')
        print('Available models are sw_LISA, sw_HL, sw_HLnew, sw_SSM, sw_LISAold')
        return 0

    OmGW  = np.zeros((len(s), len(vws), len(alphas), len(betas)))
    freqs = np.zeros((len(s), len(vws), len(betas)))
    for i in range(0, len(vws)):
        lf = hb.Rstar_beta(vws[i], cs2=cs2, corr=corrRs)/betas
        for l in range(0, len(betas)):
            # express freqs as f/H_ast (instead of f/R_ast)
            freqs[:, i, l] = s/lf[l]
            for j in range(0, len(alphas)):
                OmGW[:, i, j, l] = 3*ampl[i, j]*pref[i, j, l]*S[:, i, j]

    if redshift:

        freqs, OmGW = cGW.shift_OmGW_today(freqs, OmGW, g=gstar, gS=gS,
                                           T=T, h0=h0,  kk=False, Neff=Neff)

    if not mult_vws:
        if not mult_beta:
            freqs = freqs[:, 0, 0]
            if not mult_alpha: OmGW = OmGW[:, 0, 0, 0]
            else:              OmGW = OmGW[:, 0, :, 0]
        else:
            freqs = freqs[:, 0, :]
            if not mult_alpha: OmGW = OmGW[:, 0, 0, :]
            else:              OmGW = OmGW[:, 0, :, :]
    else:
        if not mult_beta:
            freqs = freqs[:, :, 0]
            if not mult_alpha: OmGW = OmGW[:, :, 0, 0]
            else:              OmGW = OmGW[:, :, :, 0]
        else:
            if not mult_alpha: OmGW = OmGW[:, :, 0, :]

    return freqs, OmGW

############################ TURBULENCE ############################

################# fit for the anisotropic stresses #################

def pPi_fit(s, b=b_turb, alpPi=alpPi, fPi=fPi, bPi=bPi_vort):

    """
    Function that computes the fit of the spectrum of the
    anisotropic stresses.

    The spectrum can be computed numerically for a Gaussian
    source using EPi_correlators in GW_models module.

    Default values are valid for a purely vortical velocity or
    magnetic field following a von Kárman spectrum, as indicated
    in RoperPol:2023bqa, equation 17.

    Using different values of alpPi, fPi, bPi can be generalized
    to other sources, see RoperPol:2025b

    It assumes that the anisotropic stresses can
    be expressed with the following fit:

    p_Pi = (1 + (f/fPi)^alpPi)^(-(b + bPi)/alpPi)

    Arguments:
        s     -- array of frequencies, normalized by the characteristic scale,
                 s = f R*
        b     -- high-f slope f^(-b)
        alpPi -- smoothness parameter of the fit
        fPi   -- position of the fit break
        bPi   -- extra power law decay of the spectrum of the
                 stresses compared to b

    Returns:
        Pi    -- array of the anisotropic stresses spectrum
        fGW   -- maximum value of the function s * Pi that determines
                 the amplitude of the GW spectrum for MHD turbulence
        pimax -- maximum value of Pi when s = fGW
    """

    Pi    = an.smoothed_bPL(s, a=0, b=b + bPi, kpeak=fPi,
                            alp=alpPi, norm=False, alpha2=True,
                            dlogk=False)
    pimax = ((b + bPi)/(b + bPi - 1))**(-(b + bPi)/alpPi)
    fGW   = fPi/(b + bPi - 1)**(1/alpPi)

    return Pi, fGW, pimax

def ampl_GWB_turb(a_turb=a_turb, b_turb=b_turb, alp=alp_turb):

    """
    Reference for turbulence is RoperPol:2023bqa, equation 9,
    based on the template of RoperPol:2022iel, section 3 D.

    See footnote 3 of RoperPol:2023bqa for clarification
    (extra factor 1/2 has been added to take into account average over
    oscillations that were ignored in RoperPol:2022iel).

    Arguments:
        a_turb, b_turb, alp_turb -- slopes and smoothness of the turbulent
                     source spectrum (either magnetic or kinetic), default
                     values are for a von Karman spectrum
    """

    A    = an.calA(a=a_turb, b=b_turb, alp=alp_turb)
    C    = an.calC(a=a_turb, b=b_turb, alp=alp_turb, tp='vort')
    ampl = .5*C/A**2

    return ampl

def pref_GWB_turb(Oms=Oms_ref, lf=lf_ref):

    '''
    Dependence of the GW spectrum from turbulence on the turbulence length
    scale, defined as lf = 2pi/kf, where kf is the position of the spectral peak,
    and the fraction of turbulent to radiation energy density Oms.

    Reference is RoperPol:2023bqa, equation 9, based on RoperPol:2022iel,
    section II D.

    Also used in EPTA:2023xxk, Caprini:2024hue
    See further details in RoperPol:2025b

    Arguments:
        Oms -- energy density of the source (i.e., 1/2 vrms^2)
        lf  -- mean-size of the bubbles, given as a fraction of the Hubble radius
    '''

    if not mult_Oms: Oms = [Oms]
    if not mult_lf:  lf  = [lf]
    Oms, lf = np.meshgrid(Oms, lf, indexing='ij')
    pref    = (Oms*lf)**2

    if   not mult_Oms and not mult_lf: pref = pref[0, 0]
    elif not mult_Oms: pref = pref[0, :]
    elif not mult_lf:  pref = pref[:, 0]

    return pref

def Sf_shape_turb(s, Oms=Oms_ref, lf=lf_ref, N=N_turb, cs2=cs2_ref, expansion=True,
                  tdecay=tdecay_ref, tp='magnetic', b_turb=b_turb, alpPi=alpPi,
                  fPi=fPi, bPi=bPi_vort):

    """
    Function that computes the spectral shape derived for GWs generated by
    MHD turbulence.

    Reference for vortical (MHD) turbulence is RoperPol:2023bqa,
    equation 9, based on the analytical model presented in RoperPol:2022iel,
    section II D.

    Also used in EPTA:2023xxk, Caprini:2024hue
    See further details in RoperPol:2025b

    Arguments:
        s      -- normalized wave number, divided by the mean bubbles
                  size, s = f R*
        Oms    -- energy density of the source (i.e., 1/2 vrms^2)
        lf     -- characteristic scale of the turbulence as a
                  fraction of the Hubble radius, R* H*
        N      -- relation between the decay time and the effective
                  source duration
        cs2    -- square of the speed of sound (default is 1/3)
        tdecay -- determines the finite duration in the cit model
                  (default is to use eddy turnover time)
        tp     -- type of source (default is 'magnetic', other options are
                  'kinetic' or 'max') used to compute the characteristic
                  velocity in the eddy turnover time
        b_turb -- slope of the velocity/magnetic field spectrum in the UV
        alpPi, fPi, bPi -- parameters of the pPi_fit

    Returns:
        S      -- spectral shape
    """

    if not mult_Oms: Oms = [Oms]
    if not mult_lf:  lf  = [lf]

    TGW = mod.TGW_func(s, Oms=Oms, lf=lf, N=N, cs2=cs2, expansion=expansion,
                       tdecay=tdecay, tp=tp)

    Pi, _, _ = pPi_fit(s, b=b_turb, alpPi=alpPi, fPi=fPi, bPi=bPi)
    s3Pi  = s**3*Pi
    s3Pi, Oms, lf = np.meshgrid(s3Pi, Oms, lf, indexing='ij')
    S = s3Pi/lf**2*TGW

    if   not mult_Oms and not mult_lf: S = S[:, 0, 0]
    elif not mult_Oms: S = S[:, 0, :]
    elif not mult_lf:  S = S[:, :, 0]

    return S

def OmGW_spec_turb(s, Oms, lfs, N=N_turb, cs2=cs2_ref, quiet=True, a_turb=a_turb,
                   b_turb=b_turb, alp=alp_turb, expansion=True, tdecay=tdecay_ref,
                   tp='magnetic', alpPi=alpPi, fPi=fPi, bPi=bPi_vort,
                   redshift=False, gstar=gref, gS=0, T=Tref, h0=1., Neff=Neff_ref):

    '''
    Function that computes the GW spectrum (normalized to radiation
    energy density within RD era) for turbulence.

    The general shape of the GW spectrum is based on that of reference
    RoperPol:2023bqa, equations 3 and 9:

        OmGW (f) = 3 * ampl_GWB * pref_GWB * S(f),

    where:

    - ampl_GWB is the efficiency of GW production by the specific source,
    - pref_GWB is the dependence of the GW amplitude on the source
      parameters (e.g. length scale and strength of the source),
    - S(f) is a normalized spectral shape, such that int S(f) d ln f = 1

    The GW spectrum as an observable at present time is then computed using

        OmGW0 (f) = OmGW x FGW0,

    where FGW0 is the redshift from the time of generation to present
    time, computed in cosmoGW.py that depends on the degrees of freedom at
    the time of generation.

    Arguments:
        s               -- normalized wave number, divided by the mean bubbles
                           size Rstar, s = f R*
        Oms             -- energy density of the source (i.e., 1/2 vrms^2)
        lf              -- characteristic scale of the turbulence as a
                           fraction of the Hubble radius, R* H*
        N               -- relation between the decay time and the effective
                           source duration
        cs2             -- square of the speed of sound
                           (default is 1/3 for radiation domination)
        quiet           -- option to avoid printing debugging info (default is True)
        a_turb, b_turb, -- slopes and smoothness of the turbulent
        alp_turb           source spectrum (either magnetic or kinetic), default
                           values are for a von Karman spectrum
        expansion       -- option to include the Universe expansion
                           (in radiation domination)
        tdecay          -- determines the finite duration in the cit model
                           (default is to use eddy turnover time)
        tp              -- type of source (default is 'magnetic', other options are
                           'kinetic' or 'max') used to compute the characteristic
                           velocity in the eddy turnover time
        alpPi, fPi, bPi -- parameters of the pPi_fit
        redshift        -- option to redshift the GW spectrum and frequencies to
                           present time (default is False)
        gstar           -- number of degrees of freedom (default is 100) when
                           redshift = True
        gS              -- number of adiabatic degrees of freedom (default is gstar)
                           when redshift = True
        T               -- temperature scale (default is 100 GeV) when redshift = True
        h0              -- value of the Hubble rate at present time (100 h0 Mpc/km/s)
                           (default is one)
        Neff            -- effective number of neutrino species (default is 3)

    Returns:
        freq            -- array of frequencies (normalized with Rstar if redshift
                           is False and in Hz otherwise)
        OmGW            -- GW spectrum normalized to the radiation energy density
                           (or to the present critical density if redshift is True)
    '''

    mult_Oms = isinstance(Oms, (list, tuple, np.ndarray))
    mult_lfs = isinstance(lfs,  (list, tuple, np.ndarray))

    if not mult_Oms: Oms = np.array([Oms])
    if not mult_lfs: lfs = np.array([lfs])

    #### Computing ampl_GWB

    ampl = ampl_GWB_turb(a_turb=a_turb, b_turb=b_turb, alp=alp_turb)

    #### Computing pref_GWB

    pref = pref_GWB_turb(Oms=Oms, lf=lfs)

    #### Computing the spectral shape

    S = Sf_shape_turb(s, Oms=Oms, lf=lfs, N=N, cs2=cs2, expansion=expansion,
                      tdecay=tdecay, tp=tp, b_turb=b_turb, alpPi=alpPi,
                      fPi=fPi, bPi=bPi)

    OmGW  = np.zeros((len(s), len(Oms), len(lfs)))
    freqs = np.zeros((len(s), len(lfs)))

    for l in range(0, len(lfs)):
        # express freqs as f/H_ast (instead of f/R_ast)
        freqs[:, l] = s/lf[l]
        for j in range(0, len(Oms)):
            OmGW[:, j, l] = 3*ampl*pref[j, l]*S[:, j, l]

    if redshift:
        freqs, OmGW = cGW.shift_OmGW_today(freqs, OmGW, g=gstar, gS=gS,
                                           T=T, h0=h0,  kk=False, Neff=Neff)

    if not mult_lf:
        freqs = freqs[:, 0]
        if not mult_Oms: OmGW = OmGW[:, 0, 0]
        else:            OmGW = OmGW[:, :, 0]
    else:
        if not mult_Oms: OmGW = OmGW[:, 0, :]

    return freqs, OmGW



"""
cosmoGW.py is a Python routine that contains functions relevant for
cosmological stochastic gravitational wave backgrounds (SGWB).

Adapted from the original cosmoGW in GW_turbulence
(https://github.com/AlbertoRoper/GW_turbulence),
created in Dec. 2021

Currently part of the cosmoGW code:

https://github.com/cosmoGW/cosmoGW/
https://github.com/cosmoGW/cosmoGW/blob/main/src/cosmoGW/cosmoGW.py

Author:  Alberto Roper Pol
Created: 01/12/2021 (GW_turbulence)
Updated: 13/03/2025 (release cosmoGW 1.0: https://pypi.org/project/cosmoGW)

Main references used in this library are:

Maggiore:1999vm  - M. Maggiore, "Gravitational wave experiments and
early universe cosmology," Phys.Rept. 331 (2000) 283-367,
arXiv:gr-qc/9909001.

RoperPol:2018sap - A. Roper Pol, A. Brandenburg, T. Kahniashvili,
A. Kosowsky, S. Mandal, "The timestep constraint in solving the
gravitational wave equations sourced by hydromagnetic turbulence,"
Geophys. Astrophys. Fluid Dynamics 114, 1, 130 (2020),
arXiv:1807.05479.

RoperPol:2021xnd - A. Roper Pol, S. Mandal, A. Brandenburg,
T. Kahniashvili, "Polarization of gravitational waves from helical
MHD turbulent sources," JCAP (2022) 04, 019, arXiv:2107.05356.

RoperPol:2022iel - A. Roper Pol, C. Caprini, A. Neronov, D. Semikoz,
"The gravitational wave signal from primordial magnetic fields in the
Pulsar Timing Array frequency band," Phys. Rev. D 105, 123502 (2022),
arXiv:2201.05630.
"""

import astropy.units as u
import numpy as np

def find_path():

    ## find the directory where cosmoGW is installed within sys.path
    import sys
    import os
    found = False
    paths = sys.path
    for path in paths:
      subdirs = os.walk(path)
      subdirs = list(subdirs)
      for j in subdirs:
        if 'cosmoGW' in j[0]:
            pth = j[0]
            found = True
            break
      if found: break
    return pth + '/'

HOME = find_path()

import cosmoGW.cosmology as co

# reference values and constants (EWPT)
Tref     = 100*u.GeV      # EWPT
gref     = 100            # EWPT
Neff_ref = co.Neff_ref    # reference Neff = 3

########################## SGWB at present time ##########################

def fac_hc_OmGW(d=1, h0=1.):

    """
    Function that returns the factor to transform the strain function
    hc(f) to the GW energy density OmGW (f) away from the source.

    Arguments:
        d  -- option to give the factor to convert from energy density to
             strain if set to -1 (default 1)
        h0 -- parameterizes the uncertainties (Hubble tension) in the value
              of the Hubble rate (default 1)

    Returns:
        fac -- factor to convert from the strain function hc(f) to the GW
               energy density OmGW (f) in frequency units (Hz)

    Reference: Maggiore:1999vm, eq. 17
    """

    # compute values at present day
    fac = co.H0_ref*h0*np.sqrt(3/2)/np.pi
    if d == -1: fac = 1/fac**2

    return fac

def check_frequency(f, func=''):

    try:
        f = f.to(u.Hz)
    except:
        print('Error: the input frequency in ', func,
              ' should be given in frequency units',
              ' using astropy.units, setting f to 1/year')
        f = 1/u.yr
        f = f.to(u.Hz)

    return f

def hc_OmGW(f, OmGW, d=1, h0=1.):

    """
    Function that transforms the  GW energy density OmGW (f) to the
    characteristic strain spectrum function hc(f) away from the source.
    Careful with the different notations (can be confusing!!), see below.

    Arguments:
        f    -- frequency array (in units of frequency, e.g. Hz)
        OmGW -- GW energy density spectrum OmGW (f)
        d    -- option to convert from energy density to strain if set
                to -1 (default 1)
        h0   -- parameterizes the uncertainties (Hubble tension) in the value
                of the Hubble rate (default 1)

    Returns:
        hc -- strain spectrum

    References (comment on notations):

    Main reference here is Maggiore:1999vm, eq. 17:

    OmGW(f) = (2 pi^2)/(3 H0^2) f^2 hc^2 (f).

    hc2 (f) is then defined as the spectrum in log(f) of h+^2 + hx^2 and
    Maggiore defines Sh(f) in eq. B12 such that hc^2 (f) = 2 f Sh_Mag (f).

    Note that this is different than the notation in  RoperPol:2021xnd,
    eq. B.18, used in interferometry.py, where hc^2 (f) = f Sh^pm (f),
    such that this is

    Sh^pm (f) = 2 Sh_Mag (f).

    Hence, from hc^2 (f) we can compute Sh_Mag (f) = hc^2 (f)/2f and
    Sh^pm (f) = hc^2 (f)/f.
    """

    f   = check_frequency(f, func='hc_OmGW')
    fac = fac_hc_OmGW(d=d, h0=h0)
    hc  = fac/f*np.sqrt(OmGW)
    if d==-1: hc = fac*f**2*OmGW**2

    return hc

def check_Sf(Sf, func=''):

    try:
        Sf = Sf.to(1/u.Hz**3)
    except:
        print('Error: the input spectral density in', func,
              'should be given in 1/frequency^3 using astropy.units,',
              ' setting Sf to 1/Hz^3')
        Sf = 1./u.Hz**3

    return Sf

def hc_Sf(f, Sf, d=1):

    """
    Function that transforms the power spectral density Sf (f) to the
    characteristic strain spectrum function hc(f).

    Arguments:
        f  -- frequency array (in units of frequency, e.g. Hz)
        Sf -- power spectral density Sf (f) (in units of 1/Hz^3)
        d  -- option to convert from strain to power spectral density if set
             to -1 (default 1)

    Returns:
        hc -- strain spectrum

    Reference: RoperPol:2022iel, eq. 42
    """

    f  = check_frequency(f, func='hc_Sf')
    Sf = check_Sf(Sf, func='hc_Sf')
    hc = np.sqrt(12*np.pi**2*Sf*f**3)
    if d==-1: hc = Sf**2/12/np.pi**2/f**3

    return hc

def Omega_A(A=1., fref=0, beta=0, h0=1.):

    """
    Function that returns the amplitude of the SGWB energy density
    spectrum, expressed as a power law (PL), given the amplitude A of the
    characteristic strain, also expressed as a PL.

    Note that A is always given for the reference frequency of 1/(1 year)
    and it is used in the common process reported by PTA collaborations.

    The GW energy density and characteristic amplitude can be expressed as:

    OmGW = Omref (f/fref)^beta
    hc = A (f/fyr)^alpha

    Arguments:
        A    -- amplitude of the characteristic strain PL using 1yr as the
                reference frequency
        fref -- reference frequency used for the PL expression of the
                GW background given in units of frequency (default
                1 yr^(-1))
        beta -- slope of the PL
        h0   -- parameterizes the uncertainties (Hubble tension) in the value
                of the Hubble rate (default 1)

    Returns:
        Omref -- amplitude of the GW energy density PL

    Reference: RoperPol:2022iel, eq. 44
    """

    fac = fac_hc_OmGW(d=-1, h0=h0)
    fyr = (1./u.yr).to(u.Hz)
    Omref = fac*fyr**2*A**2
    if fref != 0:
        fref = check_frequency(fref, func='Omega_A')
        Omref *= (fref.value/fyr.value)**beta

    return Omref

################ SGWB from the radiation-dominated era ################

def shift_onlyOmGW_today(OmGW, g=gref, gS=0., d=1, h0=1., Neff=Neff_ref):

    """
    Function that shifts the GW energy density spectrum from the time of
    generation to the present time (assumed to be within the RD era)

    Arguments:
        OmGW -- GW energy density spectrum per logarithmic interval
                (normalized by the radiation energy density)
        g    -- number of relativistic degrees of freedom (dof) at the time of generation
                (default is 100)
        gS   -- number of adiabatic dof (default is gS = g)
        d    -- option to reverse the transformation if set to -1, i.e.,
                to return OmGW(k) from the shifted to present time OmGW(f)
                (default 1)
        h0   -- parameterizes the uncertainties (Hubble tension) in the value
                of the Hubble rate (default 1)
        Neff -- effective number of neutrino species (default is 3)


    Returns:
        OmGW0 -- shifted spectrum OmGW to present time

    Reference: RoperPol:2022iel, eq. 27
    """

    Hs_f = co.Hs_fact()*u.MeV**2
    as_f = co.as_fact(Neff=Neff)/u.MeV
    # if gS is not specified, it is assumed to be equal to g*
    if gS==0: gS = g
    OmGW_f = Hs_f**2/co.H0_ref**2/h0**2*as_f**4*g/gS**(4./3.)
    if d==1:  OmGW0 = OmGW*OmGW_f
    if d==-1: OmGW0 = OmGW/OmGW_f

    return OmGW0

def shift_frequency_today(k, g=gref, gS=0., T=Tref, d=1, kk=True,
                          Neff=Neff_ref):

    """
    Function that transforms the normalized wave number at the time of
    generation by the Hubble rate H_* to the present time frequency.

    Arguments:
        k    -- array of wave numbers (normalized by the Hubble scale)
        g    -- number of relativistic degrees of freedom (dof) at the time of generation
                (default is 100)
        gS   -- number of adiabatic dof (default is gS = g)
        T    -- temperature scale at the time of generation in energy units
                (convertible to MeV) (default is 100 GeV)
        d    -- option to reverse the transformation if set to -1, i.e.,
                to return the normalized k from the frequency shifted to present
                time f (default 1)
        kk   -- if kk is True, then kf corresponds to k_* HH_*, otherwise it refers
                to the length in terms of the Hubble size HH_* l_* = 2 pi/(k_*/HH_*)
        Neff -- effective number of neutrino species (default is 3)

    Returns:
        f -- shifted wave number to frequency as a present time observable
             (in Hz)

    Reference: RoperPol:2022iel, eq. 32
    """

    # if gS is not specified, it is assumed to be equal to g*
    if gS==0: gS = g
    HHs = co.Hs_val(g=g, T=T)*co.as_a0_rat(g=gS, T=T, Neff=Neff)
    if d == 1:
        if kk == False: k = 2*np.pi*k
        f = k*HHs/2/np.pi
    if d==-1:
        f = 2*np.pi*k.to(u.Hz)/HHs
        if kk == False: f = 2*np.pi/f

    return f

def shift_OmGW_today(k, OmGW, g=gref, gS=0., T=Tref, d=1, h0=1.,
                     kk=True, Neff=Neff_ref):

    """
    Function that shifts the GW energy density spectrum from the time of
    generation to the present time. It assumes that the time of generation
    is within the radiation dominated era.

    Arguments:
        k    -- array of wave numbers (normalized by the Hubble scale)
        OmGW -- GW energy density spectrum per logarithmic interval
                (normalized by the radiation energy density)
        g    -- number of relativistic degrees of freedom (dof) at the time of generation
                (default is 100)
        gS   -- number of adiabatic dof (default is gS = g)
        T    -- temperature scale at the time of generation in energy units
                (convertible to MeV) (default is 100 GeV)
        d    -- option to reverse the transformation if set to -1, i.e.,
                to return the normalized k from the frequency shifted to present
                time f (default 1)
        h0   -- parameterizes the uncertainties (Hubble tension) in the value
                of the Hubble rate (default 1)
        kk   -- if kk is True, then kf corresponds to k_* HH_*, otherwise it refers
                to the length in terms of the Hubble size HH_* l_* = 2 pi/(k_*/HH_*)
        Neff -- effective number of neutrino species (default is 3)

    Returns:
        f     -- shifted wave number to frequency as a present time observable
                 (in Hz)
        OmGW0 -- shifted spectrum OmGW to present time

    Reference: see functions shift_onlyOmGW_today and shift_frequency_today
    """

    # shift Omega_GW
    OmGW0 = shift_onlyOmGW_today(OmGW, g=g, gS=gS, d=d, h0=h0, Neff=Neff)
    # shift frequency
    f     = shift_frequency_today(k,   g=g, gS=gS, T=T, d=d, kk=kk, Neff=Neff)

    return f, OmGW0

def shift_hc_today(k, hc, g=gref, gS=0., T=Tref, d=1, kk=True, Neff=Neff_ref):

    """
    Function that shifts the characteristic amplitude spectrum from the time
    of generation to the present time.

    It assumes that the time of generation is within the radiation dominated
    era.

    Arguments:
        k    -- array of wave numbers (normalized by the Hubble scale)
        hc   -- spectrum of GW characteristic amplitude per logarithmic interval
        g    -- number of relativistic degrees of freedom (dof) at the time of generation
                (default is 100)
        gS   -- number of adiabatic dof (default is gS = g)
        T    -- temperature scale at the time of generation in energy units
                (convertible to MeV) (default is 100 GeV)
        d    -- option to reverse the transformation if set to -1, i.e.,
                to return the normalized k from the frequency shifted to present
                time f, and hc(k) from the shifted to present time hc0(f) (default 1)
        kk   -- if kk is True, then kf corresponds to k_* HH_*, otherwise it refers
                to the length in terms of the Hubble size HH_* l_* = 2 pi/(k_*/HH_*)
        Neff -- effective number of neutrino species (default is 3)

    Returns:
        f    -- shifted wave number to frequency as a present time observable
                (in Hz)
        hc0  -- shifted hc spectrum to present time

    Reference: RoperPol:2018sap, eq. B.12
    """

    as_f = co.as_fact(Neff=Neff)
    T    = co.check_temperature_MeV(T, func='cosmoGW.shift_hc_today')
    hc0  = hc*as_f*g**(-1/3)/T
    if d == -1: hc0 = hc/as_f/g**(-1/3)*T
    f    = shift_frequency_today(k, g=g, gS=gS, T=T, d=d, kk=kk, Neff=Neff)

    return f, hc0



"""
GW_analytical.py is a Python routine that contains analytical
calculations and useful mathematical functions.

Adapted from the original GW_analytical in cosmoGW
(https://github.com/AlbertoRoper/cosmoGW),
created in Dec. 2021

Currently part of the cosmoGW code:

https://github.com/cosmoGW/cosmoGW/
https://github.com/cosmoGW/cosmoGW/blob/main/src/cosmoGW/GW_analytical.py

Author: Alberto Roper Pol
Created: 01/12/2021
Updated: 31/08/2024
Updated: 04/06/2025 (release cosmoGW 1.0: https://pypi.org/project/cosmoGW)

Other contributors: Antonino Midiri, Madeline Salomé

Main references are:

RoperPol:2022iel - A. Roper Pol, C. Caprini, A. Neronov, D. Semikoz,
"The gravitational wave signal from primordial magnetic fields in the
Pulsar Timing Array frequency band," Phys. Rev. D 105, 123502 (2022),
arXiv:2201.05630

RoperPol:2023bqa - A. Roper Pol, A. Neronov, C. Caprini, T. Boyer,
D. Semikoz, "LISA and γ-ray telescopes as multi-messenger probes of a
first-order cosmological phase transition," arXiv:2307.10744 (2023)

RoperPol:2023dzg - A. Roper Pol, S. Procacci, C. Caprini,
"Characterization of the gravitational wave spectrum from sound waves within
the sound shell model," Phys. Rev. D 109, 063531 (2024), arXiv:2308.12943.

Caprini:2024gyk  - A. Roper Pol, I. Stomberg, C. Caprini, R. Jinno,
T. Konstandin, H. Rubira, "Gravitational waves from first-order
phase transitions: from weak to strong," JHEP, arxiv:2409.03651

Caprini:2024hue  - E. Madge, C. Caprini, R. Jinno, M. Lewicki,
M. Merchand, G. Nardini, M. Pieroni, A. Roper Pol, V. Vaskonen,
"Gravitational waves from first-order phase transitions in LISA:
reconstruction pipeline and physics interpretation,"
JCAP 10, 020 (2024), arxiv:2403.03723

RoperPol:2025b   - A. Roper Pol, A. Midiri, M. Salomé, C. Caprini,
"Modeling the gravitational wave spectrum from slowly decaying sources in the
early Universe: constant-in-time and coherent-decay models," in preparation
"""

import numpy as np
import matplotlib.pyplot as plt

### Reference slopes
a_ref   = 5    # Batchelor spectrum k^5
b_ref   = 2/3  # Kolmogorov spectrum k^(-2/3)
alp_ref = 2    # reference smoothness of broken power-law transition

####### ANALYTICAL FUNCTIONS USED FOR A SMOOTHED BROKEN POWER LAW #######

'''
    RoperPol:2022iel and RoperPol:2025b consider a spectral function
    defined such that the average squared field corresponds to

    <v^2> ~ 2 E* k* int zeta(K) dK,     in RoperPol:2022iel
    <v^2> ~ 2 E*    int zeta(K) dlnK,   in RoperPol:2025b

    The first convention can be chosen in the following functions if
    dlogK is set to False, while the second one is assumed when dlogK
    is True
'''

def check_slopes(a, b, dlogk=True):

    '''
    Function that checks the slopes used in power spectra to ensure
    that the integral over k or log k is converging.

    Arguments:
        a -- slope of the spectrum at low wave numbers, k^a
        b -- slope of the spectrum at high wave numbers, k^(-b)
        dlogk -- option to consider the spectral function per unit of k
                 or unit of logarithm of k (see comment above)
    '''

    conv = True
    if (dlogk):
        if b < 0 or a < 0:
            print('a and b have to be larger than 0')
            conv = False
    else:
        if b < 1 or a < -1:
            print('b has to be larger than 1 and a larger than -1')
            conv = False

    return conv

def smoothed_bPL(k, A=1., a=a_ref, b=b_ref, kpeak=1., alp=alp_ref, norm=True,
                 Omega=False, alpha2=False, piecewise=False, dlogk=True):

    """
    Function that returns the value of the smoothed broken power law (bPL) model
    for a spectrum of the form:

        zeta(K) = A x (b + abs(a))^(1/alp) K^a/[ b + c K^(alp(a + b)) ]^(1/alp),

    where K = k/kpeak, c = 1 if a = 0 or c = abs(a) otherwise.

    This spectrum is defined such that kpeak is the correct position of
    the peak and its maximum amplitude is given by A when norm is True.

    If norm is set to False, then the non-normalized spectrum is used:

        zeta (K) = A x K^a/(1 + K^(alp(a + b)))^(1/alp)

    Introduced in RoperPol:2022iel, see equation 6
    Main reference is RoperPol:2025b

    Arguments:

        k      -- array of wave numbers
        A      -- amplitude of the spectrum
        a      -- slope of the spectrum at low wave numbers, k^a
        b      -- slope of the spectrum at high wave numbers, k^(-b)
        kpeak  -- spectral peak, i.e., position of the break from k^a to k^(-b)
        alp    -- smoothness of the transition from one power law to the other
        norm   -- option to normalize the spectrum such that its peak is located at
                kpeak and its maximum value is A
        Omega  -- option to use the integrated energy density as the input A
        alpha2 -- option to use the alternative convention, such that the spectrum
                  takes the form: zeta(K) ~ K^a/( b + c K^alp )^((a + b)/alp)
        piecewise -- option to return a piecewise broken power law:
                     zeta(K) = K^a for K < 1, and K^(-b) for K > 1
                     corresponding to the alpha -> infinity limit
        dlogk  -- option to consider the spectral function per unit of k
                 or unit of logarithm of k (see comment above)

    Returns:
        spec -- spectrum array
    """

    conv = check_slopes(a, b, dlogk=dlogk)
    if not conv: return 0*k**0

    c = abs(a)
    if a == 0: c = 1
    if alpha2: alp = alp/(a + b)

    K = k/kpeak
    spec = A*K**a
    if piecewise:
        spec[np.where(K > 1)] = A*K[np.where(K > 1)]**(-b)
    else:
        alp2 = alp*(a + b)
        if norm:
            m = (b + abs(a))**(1/alp)
            spec = m*spec/(b + c*K**alp2)**(1/alp)

        else: spec = spec/(1 + K**alp2)**(1/alp)

    if Omega: spec = spec/kpeak/calA(a=a, b=b, alp=alp, norm=norm,
                                     alpha2=alpha2, piecewise=piecewise,
                                     dlogk=dlogk)

    return spec

def complete_beta(a, b):

    '''
    Function that computes the complete beta function, only converges for
    positive arguments.

    B(a, b; x -> infinity) = int_0^x u^(a - 1) (1 - u)^(b - 1) du

    Arguments:
        a, b -- arguments a, b of the complete beta function

    Returns:
        B -- value of the complete beta function
    '''

    import math as m

    if a > 0 and b > 0: B = m.gamma(a)*m.gamma(b)/m.gamma(a + b)
    else:
        print('arguments of beta function need to be positive')
        B = 0

    return B

def calIab_n_alpha(a=a_ref, b=b_ref, alp=alp_ref, n=0, norm=True):

    '''
    Function that computes the normalization factor that enters in the
    calculation of Iabn

    Arguments:
        a, b -- slopes of the smoothed_bPL function
        alp  -- smoothness parameter of the smoothed_bPL function
        n    -- n-moment of the integral
        norm -- option to normalize the spectrum such that its peak is located at
                kpeak and its maximum value is 1

    Returns:
        calI -- normalization parameter that appears in the integral

    Reference: appendix A of RoperPol:2025b
    '''

    alp2 = 1/alp/(a + b)
    a_beta = (a + n + 1)*alp2

    c = abs(a)
    if a == 0: c = 1

    calI = alp2
    if norm: calI = calI*((b + abs(a))/b)**(1/alp)/(c/b)**a_beta

    return calI

def Iabn(a=a_ref, b=b_ref, alp=alp_ref, n=0, norm=True, alpha2=False,
         piecewise=False):

    '''
    Function that computes the moment n of the smoothed bPL spectra,
    defined in smoothed_bPL for A = 1, kpeak = 1:

    int K^n zeta(K) dK

    Reference: appendix A of RoperPol:2025b

    Arguments:

        a   -- slope of the spectrum at low wave numbers, k^a
        b   -- slope of the spectrum at high wave numbers, k^(-b)
        alp -- smoothness of the transition from one power law to the other
        n   -- moment of the integration

    Returns: value of the n-th moment
    '''

    if a + n + 1 <= 0:

        print('a + n has to be larger than -1 for the integral',
              'to converge')
        return 0

    if b - n - 1 <= 0:

        print('b + n has to be larger than 1 for the integral',
              'to converge')
        return 0

    if piecewise:

        return (a + b)/(a + n + 1)/(b - n - 1)

    if alpha2: alp = alp/(a + b)
    alp2 = 1/alp/(a + b)
    a_beta = (a + n + 1)*alp2
    b_beta = (b - n - 1)*alp2
    calI = calIab_n_alpha(a=a, b=b, alp=alp, n=n, norm=norm)
    comp_beta = complete_beta(a_beta, b_beta)

    return comp_beta*calI

def calA(a=a_ref, b=b_ref, alp=alp_ref, norm=True, alpha2=False,
         piecewise=False, dlogk=True):

    '''
    Function that computes the parameter calA = Iab,0 that relates the
    peak and the integrated values of the smoothed_bPL spectrum

    References are RoperPol:2022iel, equation 8 (for dlogk False)
    and RoperPol:2025b, appendix A (for dlogk True)

    Same arguments as Iabn function with n = 0 or -1
    '''

    nn = 0
    if dlogk: nn = -1

    return Iabn(a=a, b=b, alp=alp, n=nn, norm=norm, alpha2=alpha2,
                piecewise=piecewise)

def calB(a=a_ref, b=b_ref, alp=alp_ref, n=1, norm=True, alpha2=False,
         piecewise=False, dlogk=True):

    '''
    Function that computes the parameter calB = Iab;-1/Iab;0 that relates the
    peak and the integral scale, cal B = xi kpeak

    Same arguments as Iabn function with n = -1 (Im1) and n = 0 (I0)
    Returns calB = (Im1/Im0)^n

    Main reference is RoperPol:2025b, appendix A (for dlogk True)
    '''

    nn = 0
    if dlogk: nn = -1

    Im1 = Iabn(a=a, b=b, alp=alp, n=-n + nn, norm=norm, alpha2=alpha2,
               piecewise=piecewise)
    I0 = Iabn(a=a, b=b, alp=alp, n=nn, norm=norm, alpha2=alpha2,
              piecewise=piecewise)
    calB = (Im1/I0)**n

    return calB

def zetam(a=a_ref, b=b_ref, alp=alp_ref, m=-1, norm=True, alpha2=False,
          piecewise=False):

    '''
    Function that computes the amplitude at the peak of the function

    K^m zeta(K)

    It allows to relate spectral functions using different normalizations,
    e.g., m = -1 allows to relate zeta used in RoperPol:2022iel with
    zeta used in RoperPol:2025b

    Main reference is RoperPol:2025b, appendix A

    Arguments:
        a      -- slope of the spectrum at low wave numbers, k^a
        b      -- slope of the spectrum at high wave numbers, k^(-b)
        kpeak  -- spectral peak, i.e., position of the break from k^a to k^(-b)
        alp    -- smoothness of the transition from one power law to the other
        m      -- power of K^m used to compute new function K^m zeta(K)
        norm   -- option to normalize the spectrum such that its peak is located at
                        kpeak and its maximum value is A
        alpha2 -- option to use the alternative convention, such that the spectrum
                          takes the form: zeta(K) ~ K^a/( b + c K^alp )^((a + b)/alp)
        piecewise -- option to return a piecewise broken power law:
                             zeta(K) = K^a for K < 1, and K^(-b) for K > 1
                             corresponding to the alpha -> infinity limit

    Returns:
        zetam -- amplitude of new function K^m zeta(K)
    '''

    if piecewise: return 1.

    if alpha2: alp2 = alp
    else: alp2 = alp*(a + b)
    zetam = ((a + m)/a)**((a + m)/alp2)*((b - m)/b)**((b - m)/alp2)

    return zetam

def Kstarm(a=a_ref, b=b_ref, alp=alp_ref, m=-1, norm=True, alpha2=False,
           piecewise=False):

    '''
    Function that computes the location of the peak of the function K^m zeta(K)

    Arguments are same as zetam
    Returns:
        Kstarm -- position of the peak of new function K^m zeta(K)
    '''

    if piecewise: return 1.

    if alpha2: alp2 = alp
    else: alp2 = alp*(a + b)
    Kstarm = (b/a*(a + m)/(b - m))**(1./alp2)

    return Kstarm

def calC(a=a_ref, b=b_ref, alp=alp_ref, tp='vort', norm=True, alpha2=False,
         piecewise=False, dlogk=True):

    '''
    Function that computes the parameter calC that allows to
    compute the TT-projected stress spectrum by taking the convolution of the
    smoothed bPL spectra over k and tilde p = |k - p|.

    It gives the value of the spectrum at K -> 0 limit as a prefactor
    pref that depends on the type of source and an integral over

    zeta^2/P^4     when dlogk is True (see appendix B of RoperPol:2025b)
    zeta^2/P^2     when dlogk is False (eq. 22 of RoperPol:2022iel for
                     vortical fields and eq. 46 of RoperPol:2023dzg
                     for compressional fields

    The spectrum of the stresses is then normalized such that
    (see appendix B of RoperPol:2025b)

    EPi = K^3 Epeak^2 calC zetaPi

    For the antisymmetric spectrum (helical) of the stresses
    the normalization becomes

    HPi = K^4 Hpeak Epeak calC zetaPi

    Arguments:

        a -- slope of the spectrum at low wave numbers, k^a
        b -- slope of the spectrum at high wave numbers, k^(-b)
        alp -- smoothness of the transition from one power law to the other
        tp -- type of sourcing field: 'vort', 'comp', 'hel' or 'mix' available
        norm -- option to normalize the spectrum such that its peak is located at
                        kpeak and its maximum value is A
        alpha2 -- option to use the alternative convention, such that the spectrum
                          takes the form: zeta(K) ~ K^a/( b + c K^alp )^((a + b)/alp)
        piecewise -- option to return a piecewise broken power law:
                             zeta(K) = K^a for K < 1, and K^(-b) for K > 1
                             corresponding to the alpha -> infinity limit
    '''

    pref = 1.

    if tp   == 'vort': pref = 28/15
    elif tp == 'comp': pref = 32/15
    elif tp == 'mix':  pref = 16/5
    elif tp == 'hel':  pref = 1/3

    if tp not in ['vort', 'comp', 'mix', 'hel', 'none']:
        print('tp has to be vortical (vort), compressional (comp),',
              'mixed (mix) or helical (hel)')
        print('returning pref = 1')

    nn = 0
    if dlogk: nn = 2

    return pref*Iabn(a=a*2, b=b*2, alp=alp/2, n=-2-nn, norm=norm,
                     alpha2=alpha2, piecewise=piecewise)

###### ANALYTICAL TEMPLATE USED FOR A DOUBLE SMOOTHED BROKEN POWER LAW ######

def smoothed_double_bPL(k, kpeak1, kpeak2, A=1., a=a_ref, b=1,
                        c=b_ref, alp1=alp_ref, alp2=alp_ref, kref=1.,
                        alpha2=False):

    """
    Function that returns the value of the smoothed double broken power
    law (double_bPL) model with a spectrum of the form:

        zeta(K) = A K^a/(1 + (K/K1)^[(a - b)*alp1])^(1/alp1)
                  x (1 + (K/K2)^[(c + b)*alp2])^(-1/alp2)

    where K = k/kref, K1 and K2 are the two position peaks,
    a is the low-k slope, b is the intermediate slope,
    and -c is the high-k slope.
    alp1 and alp2 are the smoothness parameters for each spectral transition.

    Reference is RoperPol:2023dzg, equation 50
    Also used in RoperPol:2023bqa, equation 7

    The same broken power law with a slightly different form is used in
    Caprini:2024gyk, Caprini:2024hue and can be used setting alpha2 = True

        zeta(K) = A K^a/(1 + (K/K1)^alp1)^((a - b)/alp1)
                           x (1 + (K/K2)^alp2)^(-(b + c)/alp2)

    Arguments:

        k -- array of wave numbers
        kpeak1, kpeak2 -- peak positions
        A -- amplitude of the spectrum
        a -- slope of the spectrum at low wave numbers, k^a
        b -- slope of the spectrum at intermediate wave numbers, k^b
        c -- slope of the spectrum at high wave numbers, k^(-c)
        alp1, alp2 -- smoothness of the transitions from one power law to the other
        kref   -- reference wave number used to normalize the spectrum (default is 1)
        alpha2 -- option to use different normalization

    Returns:
        spectrum array
    """

    K  = k/kref
    K1 = kpeak1/kref
    K2 = kpeak2/kref

    if not alpha2:
        spec1 = (1 + (K/K1)**((a - b)*alp1))**(1/alp1)
        spec2 = (1 + (K/K2)**((c + b)*alp2))**(1/alp2)
    else:
        spec1 = (1 + (K/K1)**alp1)**((a - b)/alp1)
        spec2 = (1 + (K/K2)**alp2)**((c + b)/alp2)
    spec = A*K**a/spec1/spec2

    return spec
